# -------------------------------------------------------------------------------
# Copyright 2006-2020 UT-Battelle, LLC. See LICENSE for more information.
# -------------------------------------------------------------------------------
"""IPS Framework Component"""
import sys
import os
import weakref
from copy import copy
from .messages import Message, MethodResultMessage


class Component:
    """
    Base class for all IPS components.  Common set up, connection and
    invocation actions are implemented here.

    :param services: service proxy to communicate with framework
    :type services: :class:`~ipsframework.services.ServicesProxy`

    :param config: configuration dictionary for this component
    :type config: dict
    """

    def __init__(self, services, config):
        """
        Set up config values and reference to services.
        """
        self.component_id = None
        self.invocation_q = None
        self.services = weakref.proxy(services)
        self.config = config
        self.start_time = 0.0
        self.sys_exit = None
        for i in config.keys():
            try:
                setattr(self, i, config[i])
            except Exception as e:
                print('Error setting Component parameter : ', i, ' - ', e)
                raise

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        for k, v in self.__dict__.items():
            if k in ["invocation_q", "sys_exit", "services"]:
                setattr(result, k, None)
            else:
                setattr(result, k, copy(v))
        return result

    def __initialize__(self, component_id, invocation_q, start_time=0.0):
        """
        Establish connection to *invocation_q*.
        """
        self.component_id = component_id
        self.invocation_q = invocation_q
        self.start_time = start_time
#        setattr(sys, 'exit', sys.exit)

    def __my_exit__(self, arg=0):
        """
        Produce message and exception about exit.
        """
        self.services.error('Called sys.exit() from component code')
        raise Exception('Called sys.exit() from component code')

    def __run__(self):
        """
        Set (and possibly create) the working directory for the component
        and change the working directory to the (possibly newly created)
        directory.
        Wait for incoming commands delivered via the *invocation_q*, and
        dispatch the incoming methods accordingly.
        """

        tmp = sys.exit
        sys.exit = self.__my_exit__
        self.sys_exit = tmp
        try:
            redirect = self.services.sim_conf['OUT_REDIRECT']
        except KeyError:
            pass
        else:
            if str(redirect).strip() != '':
                if ('OUT_REDIRECT_FNAME') not in list(self.services.sim_conf.keys()):
                    fname = "%s.out" % (self.services.sim_conf['SIM_NAME'])
                    fname = os.path.join(self.services.sim_conf['PWD'], fname)
                    print('Redirecting stdout to ', fname)
                else:
                    fname = self.services.sim_conf['OUT_REDIRECT_FNAME']
                original_stdout_fd = sys.stdout.fileno()
                original_stderr_fd = sys.stderr.fileno()
                outf = open(fname, "a")
                outf_fno = outf.fileno()
                # sys.stdout.close()
                os.dup2(outf_fno, original_stdout_fd)
                os.dup2(outf_fno, original_stderr_fd)
                # Use line buffered for stderr/stdout redirected files
                sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 1)
                sys.stderr = os.fdopen(sys.stderr.fileno(), 'w', 1)

        # SIMYAN: reversed changes that took directory creation in work out of
        # a component's hands. Now this class creates the directory and changes
        # into it as before.
        workdir = self.services.get_working_dir()

        try:
            os.chdir(workdir)
        except OSError:
            self.services.debug('Working directory %s does not exist - will attempt creation',
                                workdir)
            try:
                os.makedirs(workdir)
            except OSError as oserr:
                self.services.exception('Error creating directory %s : %s',
                                        workdir, oserr.strerror)
                raise
            os.chdir(workdir)
        self.services.debug('Running - CompID =  %s',
                            self.component_id.get_serialization())

        if self.services.profile:
            self.services.debug('Instrumenting - CompID =  %s',
                                self.component_id.get_serialization())
        self.services._init_event_service()

        while True:
            msg = self.invocation_q.get()
            self.services.log('Received Message ')
            sender_id = msg.sender_id
            call_id = msg.call_id
            method_name = msg.target_method
            args = msg.args
            keywords = msg.keywords
            formatted_args = ['%.3f' % (x) if isinstance(x, float)
                              else str(x) for x in args]
            if keywords:
                formatted_args += [" %s=" % k + str(v) for (k, v) in keywords.items()]

            self.services.debug('Calling method ' + method_name +
                                "(" + ' ,'.join(formatted_args) + ")")
            try:
                method = getattr(self, method_name)
                retval = method(*args, **keywords)
            except Exception as e:
                self.services.exception('Uncaught Exception in component method.')
                response_msg = MethodResultMessage(self.component_id,
                                                   sender_id,
                                                   call_id,
                                                   Message.FAILURE, e)
            else:
                response_msg = MethodResultMessage(self.component_id,
                                                   sender_id,
                                                   call_id,
                                                   Message.SUCCESS, retval)
            self.services.fwk_in_q.put(response_msg)

    def init(self, timestamp=0.0, **keywords):
        """
        Produce some default debugging information before the rest of the code
        is executed.
        """
        self.services.debug('init() method called')

    def restart(self, timestamp=0.0, **keywords):
        """
        Produce some default debugging information before the rest of the code
        is executed.
        """
        self.services.debug('restart() method called')

    def step(self, timestamp=0.0, **keywords):
        """
        Produce some default debugging information before the rest of the code
        is executed.
        """
        self.services.debug('step() method called')

    def finalize(self, timestamp=0.0, **keywords):
        """
        Produce some default debugging information before the rest of the code
        is executed.
        """
        self.services.debug('finalize() method called')

    def checkpoint(self, timestamp=0.0, **keywords):
        """
        Produce some default debugging information before the rest of the code
        is executed.
        """
        self.services.debug('checkpoint() method called')

    def terminate(self, status):
        """
        Clean up services and call :py:obj:`sys_exit`.
        """
        self.services.cleanup()
        if status == Message.SUCCESS:
            self.services.debug('Calling self.sys_exit(0)')
            self.sys_exit(0)
        else:
            self.services.debug('Calling self.sys_exit(1)')
            self.sys_exit(1)



"""
The Resource Helper file contains all of the code needed to figure out what
host we are on and what resources we have.  Taking this out
of the resource manager will allow us to test it independent
of the IPS.
"""

import os
import subprocess
from math import ceil
from .ipsExceptions import InvalidResourceSettingsException


def get_qstat_jobinfo():
    """
    Use ``qstat -f $PBS_JOBID`` to get the number of nodes and ppn of the
    allocation.  Typically works on PBS systems.
    """
    try:
        job_id = os.environ['PBS_JOBID']
    except Exception:
        raise

    shell_host = None
    try:
        shell_host = os.environ['HOST']
    except Exception:
        pass

    command = 'qstat -f %s' % (job_id)
    p = subprocess.Popen(command.split(), stdout=subprocess.PIPE)
    p.wait()
    if p.returncode == 0:
        out = p.stdout.readlines()
        if shell_host == 'stix':
            start_line = -1
            end_line = -1
            ppn = 0
            for i, line in enumerate(out):
                if line.strip().startswith('exec_host ='):
                    start_line = i
                if start_line != -1 and end_line == -1:
                    if '=' in line:
                        end_line = i
                if line.strip().startswith('Resource_List.nodes = '):
                    if 'ppn=' in line:
                        ppn = int(line.split('=')[-1])
                    else:
                        ppn = 8

            host_out = ''
            for i in range(start_line, end_line + 1):
                host_out += out[i]

            host_string = host_out.strip().replace(' \n', '').split('=')[1]
            node_list = [t .strip() for t in host_string.split('+')]
            num_nodes = len(node_list)
            return num_nodes, ppn, True, node_list
        else:
            width = [x.strip() for x in out if 'Resource_List.mppwidth' in x]
            mpp_npp = [x.strip() for x in out if 'Resource_List.mppnppn' in x]
            num_procs = int(width[0].split('=')[1])
            if len(mpp_npp) > 0:
                ppn = int(mpp_npp[0].split('=')[1])
            num_nodes = int(ceil(float(num_procs) / float(ppn)))
            return num_nodes, ppn, False, []
    else:
        raise Exception('Error in call to qstat.')


def get_qstat_jobinfo2():
    """
    A second way to use ``qstat -f $PBS_JOBID`` to get the number
    of nodes and ppn of the
    allocation.  Typically works on PBS systems.
    """
    try:
        job_id = os.environ['PBS_JOBID']
    except Exception:
        raise

    command = 'qstat -f %s' % (job_id)
    p = subprocess.Popen(command.split(), stdout=subprocess.PIPE)
    p.wait()
    if p.returncode == 0:
        out = p.stdout.readlines()
        found_start = False
        found_end = False
        lines = ''
        for line in out:
            if 'exec_host' in line:
                found_start = True
                lines += line.strip()
            elif 'Hold_Types' in line:
                found_end = True
            elif found_start and not found_end:
                lines += line.strip()

        lines = lines.split('=')[1].strip()
        ppn = 1
        nodes = []
        ndata = []
        for k in lines.split('+'):
            node_name, procid = k.split('/')
            if node_name not in nodes:
                nodes.append(node_name)
                ndata.append((node_name, [procid]))
            else:
                i = nodes.index(node_name)
                ndata[i][1].append(procid)
        ppn = max([len(p[1]) for p in ndata])
        return len(ndata), ppn, False, ndata
    else:
        raise Exception('Error in call to qstat')


def get_checkjob_info():
    ndata = []
    nodes = []
    cmd = "checkjob $PBS_JOBID"
    p = 0
    tot_procs = 0
    data_lines = []
    job_id = os.getenv('PBS_JOBID', '')
    mixed_nodes = False
    # Test for interactive use on batch platforms
    if job_id == '':
        raise Exception('Cannot find PBS_JOBID')
    # run checkjob $PBS_JOBID
    try:
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
        lines = [line.strip() for line in proc.stdout.readlines()]
        start = end = 0
        for k, line in enumerate(lines):
            x = line.rstrip()
            if x == "Allocated Nodes:" and start == 0:
                start = k + 1
            elif x == '' and start > 0 and end == 0:
                end = k
            if x.find("Total Requested Tasks:") > -1:
                _, b = x.split(":")
                b = b.strip()
                tot_procs = int(b)
        for line in lines[start:end + 1]:
            if line.strip() != "":
                data_lines.append(line.strip())
    except Exception as e:
        print(e)
        raise e
        # return nodes, procs
    # parse output to get allocated nodes data
    """
    There are two different formats for listing nodes that the job has access to.
    For the small numbers of nodes:
            [node_id:tasks_per_node]+

    For large numbers of nodes:
            [list of comma separated node_ids and node_id ranges]*tasks_per_node
    """
    if data_lines[0].find(":") > -1:
        # small node number format
        try:
            for j in data_lines:
                j = j[1:len(j) - 1]  # strip off first [ and last ]
                pairs = j.split('][')
                for i in pairs:
                    ndata.append(i.split(':'))
            # parse allocated nodes data [nid:nprocs]...
            for (m, p) in ndata:
                nodes.append(m)
        except Exception as e:
            print('problem parsing - small format')
            raise e
    elif data_lines[0].find("*") > -1:
        # large node number format
        try:
            nodes_str = ""
            data = ""
            # put the whole thing on one line
            data = "".join(data_lines)
            nodes_str, p = data.split("*")
            nodes_str = nodes_str.strip("[]")
            ranges = nodes_str.split(",")
            for r in ranges:
                if r.find("-") > -1:
                    # this is a range
                    ss, es = r.split("-")
                    s = int(ss)
                    e = int(es)
                    for i in range(s, e + 1):
                        nodes.append(str(i))
                else:
                    # this is a single node id
                    nodes.append(r)
            ndata = [(n, p) for n in nodes]
        except Exception as e:
            print('problem parsing - large format')
            raise e
    else:
        # TODO: make this into a real exception type
        raise Exception("could not parse resource data")
    if abs(len(nodes) * int(p) - tot_procs) > 1:
        print('len(nodes) = %d  p = %d  tot_procs = %d' % (len(nodes), int(p), tot_procs))
        print("something wrong with parsing - node count*cores does not match task count")
        raise Exception("something wrong with parsing - node count*cores does not match task count")
    return nodes, int(p), mixed_nodes, ndata


def get_slurm_info():
    """
    Access environment variables set by Slurm to get the node names,
    tasks per node and number of processes.

      ``SLURM_NODELIST``
      ``SLURM_TASKS_PER_NODE`` or ``SLURM_JOB_TASKS_PER_NODE``
      ``SLURM_NPROC``
    """
    nodes = []
    mixed_nodes = False
    try:
        nodelist = os.environ['SLURM_NODELIST']
    except Exception:
        raise
    try:
        ppn = os.environ['SLURM_TASKS_PER_NODE']
        ppn = int(ppn.split("(")[0])
    except Exception:
        try:
            ppn = os.environ['SLURM_JOB_TASKS_PER_NODE']
            ppn = int(ppn.split("(")[0])
        except Exception:
            raise
    max_p = ppn
    try:
        nproc = int(os.environ['SLURM_NPROC'])
    except Exception:
        # need to set later
        nproc = 0

    try:
        cmd = 'scontrol show hostname %s' % nodelist
        sys_nodes = subprocess.check_output(cmd.split(), encoding='UTF-8').strip().split('\n')
        nodes.extend([(k, ppn) for k in sys_nodes])
        print('IPS SLURM_NODES = ', nodes)
    except Exception:
        raise

    if 0 < nproc < len(nodes) * max_p:
        mixed_nodes = True
        nodes[-1][1] = nproc % max_p

    return len(nodes), max_p, mixed_nodes, nodes


def get_pbs_info():
    """
    Access info about allocation from PBS environment variables:

        ``PBS_NNODES``
        ``PBS_NODEFILE``
    """
    try:
        node_file = os.environ['PBS_NODEFILE']
        # core_list is a misnomer, it is a list of (repeated) node names
        # where the node names are repeated for each process they can service
        core_list_all = [line.strip() for line in open(node_file, 'r').readlines()]
        core_list = core_list_all
        node_dict = {}
        for core in core_list:
            # core is really a node name
            try:
                node_dict[core] += 1
            except KeyError:
                node_dict[core] = 1
        listOfNodes = list(node_dict.items())
        max_p = max(node_dict.values())
        mixed_nodes = (max_p != min(node_dict.values()))
        return len(listOfNodes), max_p, mixed_nodes, listOfNodes
    except Exception:
        try:
            node_count = int(os.environ['PBS_NNODES'])
            return node_count, 0, False, []
        except Exception:
            raise


def manual_detection(services):
    """
    Use values listed in platform configuration file.
    """
    listOfNodes = []
    num_nodes = int(services.get_platform_parameter('NODES'))
    ppn = int(services.get_platform_parameter('PROCS_PER_NODE'))
    tot_procs = int(services.get_platform_parameter('TOTAL_PROCS'))
    cpn = int(services.get_platform_parameter('CORES_PER_NODE'))
    if tot_procs == 0:
        if num_nodes == 0:
            if ppn == 0:
                if cpn == 0:
                    cpn = 1
                ppn = cpn
            num_nodes = 1
        tot_procs = num_nodes * ppn

    for n in range(num_nodes):
        listOfNodes.append(("dummynode%d" % n, ppn))
    if tot_procs < num_nodes * (ppn - 1):
        n = listOfNodes[-1][0]
        listOfNodes[-1] = (n, tot_procs % ppn)
    return num_nodes, ppn, False, listOfNodes


def getResourceList(services, host, partial_nodes=False):
    """
    Using the host information, the resources are detected.  Return list of
    (<node name>, <processes per node>), cores per node, sockets per node,
    processes per node, and ``True`` if the node names are accurate, ``False``
    otherwise.
    """
    listOfNodes = []
    # get the number of nodes for that machine
    num_nodes = 1
    ppn = 1
    spn = 1
    cpn = 1
    accurateNodes = False
    mixed_nodes = False

    node_detect_str = services.get_platform_parameter('NODE_DETECTION',
                                                      silent=True)
    if node_detect_str == "checkjob":
        num_nodes, ppn, mixed_nodes, listOfNodes = get_checkjob_info()
        print("=======================================================")
        print(num_nodes, ppn, mixed_nodes, listOfNodes)
        accurateNodes = False
    elif node_detect_str == "qstat":
        num_nodes, ppn, mixed_nodes, listOfNodes = get_qstat_jobinfo()
        accurateNodes = False
    elif node_detect_str == "qstat2":
        num_nodes, ppn, mixed_nodes, listOfNodes = get_qstat_jobinfo2()
        accurateNodes = True
    elif node_detect_str == "pbs_env":
        num_nodes, ppn, mixed_nodes, listOfNodes = get_pbs_info()
        if ppn == 0:
            ppn = 1
        if not listOfNodes:
            for n in range(num_nodes):
                listOfNodes.append(("dummynode%d" % n, ppn))
        else:
            accurateNodes = True
    elif node_detect_str == "slurm_env":
        num_nodes, ppn, mixed_nodes, listOfNodes = get_slurm_info()
        accurateNodes = True
    elif node_detect_str == "manual":
        num_nodes, ppn, mixed_nodes, listOfNodes = manual_detection(services)
        accurateNodes = False
    else:
        print("WARNING: no node detection strategy specified in platform config file ('NODE_DETECTION'). "
              "Valid options are: checkjob, qstat, pbs_env, slurm_env, manual.  Trying all detection schemes.")
        try:
            num_nodes, ppn, mixed_nodes, listOfNodes = get_checkjob_info()
            accurateNodes = True
        except Exception:
            try:
                num_nodes, ppn, mixed_nodes, listOfNodes = get_qstat_jobinfo()
                accurateNodes = False
            except Exception:
                try:
                    num_nodes, ppn, mixed_nodes, listOfNodes = get_pbs_info()
                    if ppn == 0:
                        ppn = 1
                    if not listOfNodes:
                        for n in range(num_nodes):
                            listOfNodes.append(("dummynode%d" % n, ppn))
                    else:
                        accurateNodes = True
                except Exception:
                    try:
                        num_nodes, ppn, mixed_nodes, listOfNodes = get_slurm_info()
                        accurateNodes = True
                    except Exception:
                        try:
                            num_nodes, ppn, mixed_nodes, listOfNodes = manual_detection(services)
                            accurateNodes = False
                        except Exception:
                            print("*** NO DETECTION MECHANISM WORKS ***")
                            raise
    # detect topology
    cpn = int(services.get_platform_parameter('CORES_PER_NODE'))
    spn = int(services.get_platform_parameter('SOCKETS_PER_NODE'))
    if cpn <= 0:
        cpn = ppn
    elif cpn < ppn:
        ppn = cpn
        if not mixed_nodes:
            for i, node in enumerate(listOfNodes):
                name = node[0]
                listOfNodes[i] = (name, ppn)
    if spn <= 0:
        spn = 1
    elif spn > cpn:
        raise InvalidResourceSettingsException("spn > cpn", spn, cpn)
    elif cpn % spn != 0:
        raise InvalidResourceSettingsException("spn not divisible by cpn", spn, cpn)
    return listOfNodes, cpn, spn, ppn, accurateNodes



# -------------------------------------------------------------------------------
# Copyright 2006-2020 UT-Battelle, LLC. See LICENSE for more information.
# -------------------------------------------------------------------------------
"""
The CCA event service, but without CCA-specific extensions:
https://www.cca-forum.org/wiki/tiki-index.php?page=Event+Specification+Proposal

It pretty much forwards calls to an event proxy which in turn talks to the
central event service.
"""
from copy import deepcopy


class EventServiceException(Exception):
    """
    Exception class for the event service.
    """

    def __init__(self, value):
        super().__init__()
        self.value = value

    def __str__(self):
        return self.value


class PublisherEventService:
    """
    Interface to topics for publishers.
    """

    def getTopic(self, topicName):
        """

        """
        return _proxy.getTopic(topicName)

    def existsTopic(self, topicName):
        return _proxy.existsTopic(topicName)


class SubscriberEventService:
    def __init__(self):
        self.subscriberid = _proxy.registerSubscriber()

    def getSubscription(self, subscriptionName):
        """
        A Subscription object can be safely returned from here without screwing
        up automatic object tracking for cleaning up out-of-scope subscriptions.
        A framework/component subscriber uses this Subscription object to
        talk to the event service.
        """
        _proxy.getSubscription(self.subscriberid, subscriptionName)
        return Subscription(self.subscriberid, subscriptionName)

    def processEvents(self):
        _proxy.processEvents(self.subscriberid)

    def __del__(self):
        _proxy.unregisterSubscriber(self.subscriberid)


class Event:
    def __init__(self, header={}, body={}):
        self.header = deepcopy(header)
        self.body = deepcopy(body)

    def getHeader(self):
        return self.header

    def getBody(self):
        return self.body

    def __str__(self):
        return str(self.body)


class EventListener:
    def __init__(self):
        self.listenerid = _proxy.createListener()

    def processEvent(self, topicName, theEvent):
        """
        A listener implements the processEvent method to respond to an event,
        thereby overriding the below invocation. Ideally, it should be an abstract
        method, but currently serves to check the correct operation of the
        event service.
        """
        # TODO: To be deleted#
        _proxy._processEvent(self.listenerid, topicName, theEvent)


class Topic:
    def __init__(self, topicName):
        self.topicName = topicName

    def getTopicName(self):
        return self.topicName

    def sendEvent(self, eventName, eventBody):
        _proxy.sendEvent(self.topicName, eventName, eventBody)


class Subscription:
    def __init__(self, subscriberid, subscriptionName):
        self.subscriberid = subscriberid
        self.subscriptionName = subscriptionName

    def registerEventListener(self, listenerKey, theListener):
        _proxy.registerEventListener(self.subscriberid, self.subscriptionName,
                                     listenerKey, theListener.listenerid, theListener)

    def unregisterEventListener(self, listenerKey):
        _proxy.unregisterEventListener(self.subscriberid, self.subscriptionName, listenerKey)

    def getSubscriptionName(self):
        return self.subscriptionName

    def __del__(self):
        _proxy.removeSubscription(self.subscriberid, self.subscriptionName)


""" Initialize the proxy """


def initialize_event_service(service):
    global _proxy

    if isinstance(service, EventService):
        _proxy = EventServiceFwkProxy(service)
    else:
        _proxy = EventServiceCmpProxy(service)


# pylint: disable=C0413
from .eventService import EventService  # noqa: E402
from .eventServiceProxy import EventServiceFwkProxy, EventServiceCmpProxy  # noqa: E402



# -------------------------------------------------------------------------------
# Copyright 2006-2020 UT-Battelle, LLC. See LICENSE for more information.
# -------------------------------------------------------------------------------
import sys
import os
import subprocess
from . import ipsutil


class DataManager:
    """
    The data manager facilitates the movement and exchange of data files for
    the simulation.
    """
    # DM init

    def __init__(self, fwk):
        # ref to framework
        self.fwk = fwk
        self.ES = None
        self.TM = None
        self.RM = None
        self.CM = None
        self.host = self.fwk.host
        self.myTopic = None
        self.outPrefix = ""
        self.simroot = ""
        self.statedir = ""
        self.state_files = []
        self.service_methods = ['stage_state',
                                'update_state',
                                'merge_current_plasma_state']
        self.fwk.register_service_handler(self.service_methods,
                                          getattr(self, 'process_service_request'))

    def process_service_request(self, msg):
        """
        Invokes the appropriate public data manager method for the component
        specified in *msg*.  Return method's return value.
        """
        self.fwk.debug('Data Manager received message: %s', str(msg))
        method = getattr(self, msg.target_method)
        retval = method(msg)
        return retval

    def stage_state(self, msg):
        """
        Copy plasma state files from source dir to target dir.  Return 0.
        Exception raised on copy error.

        *msg.args*:

          0. state_files
          1. source_dir
          2. target_dir
        """
        state_files = msg.args[0]
        source_dir = msg.args[1]
        target_dir = msg.args[2]
        try:
            ipsutil.copyFiles(source_dir, state_files, target_dir)
        except Exception:
            self.fwk.exception('Error staging plasma state files to directory %s',
                               target_dir)
            raise
        return 0

    def update_state(self, msg):
        """
        Copy plasma state files from source dir to target dir.  Return 0.
        Exception raised on copy error.

        *msg.args*:

          0. state_files
          1. source_dir
          2. target_dir
        """
        state_files = msg.args[0]
        source_dir = msg.args[1]
        target_dir = msg.args[2]
        try:
            ipsutil.copyFiles(source_dir, state_files, target_dir)
        except Exception:
            self.fwk.exception('Error updating state files from directory %s',
                               source_dir)
            raise
        return 0

    def merge_current_plasma_state(self, msg):
        """
        Merge partial plasma state file with global master.  Newly updated
        plasma state copied to caller's workdir.
        Exception raised on copy error.

        *msg.args*:

          0. partial_state_file
          1. target_state_file
          2. log_file: stdout for merge process if not ``None``
        """
        partial_state_file = msg.args[0]
        target_state_file = msg.args[1]
        log_file = msg.args[2]
        update_state = msg.args[3]

        plasma_work_dir = os.path.dirname(target_state_file)
        component_work_dir = os.path.dirname(partial_state_file)
        current_plasma_state = os.path.basename(target_state_file)

        merge_stdout = sys.stdout
        if log_file:
            log_fullpath = os.path.join(component_work_dir, log_file)
            try:
                merge_stdout = open(log_fullpath, 'w')
            except Exception:
                self.fwk.exception('Error opening log file %s : using stdout',
                                   log_fullpath)

        try:
            retval = subprocess.call([update_state, '-input', target_state_file,
                                      '-updates', partial_state_file],
                                     stdout=merge_stdout,
                                     stderr=subprocess.STDOUT)
        except Exception:
            self.fwk.exception('Error calling update_state - probably not found in $PATH')
            raise

        if retval != 0:
            return retval
        try:
            ipsutil.copyFiles(plasma_work_dir, current_plasma_state, component_work_dir)
        except Exception:
            self.fwk.exception('Error refreshing local copy of current plasma state file in directory %s',
                               component_work_dir)
            raise
        return 0



#! /usr/bin/env python3
# -------------------------------------------------------------------------------
# Copyright 2006-2020 UT-Battelle, LLC. See LICENSE for more information.
# -------------------------------------------------------------------------------
"""
   The Integrated Plasma Simulator (IPS) Framework. This framework enables
   loose, file-based coupling of certain class of nuclear fusion simulation
   codes.

   For further design information see

    - Wael Elwasif, David E. Bernholdt, Aniruddha G. Shet, Samantha S. Foley,
      Randall Bramley, Donald B. Batchelor, and Lee A. Berry, *The Design and
      Implementation of the SWIM Integrated Plasma Simulator*, in The 18th
      Euromirco International Conference on Parallel, Distributed and
      Network - Based Computing (PDP 2010), 2010.
    - Samantha S. Foley, Wael R. Elwasif, David E. Bernholdt, Aniruddha G.
      Shet, and Randall Bramley, *Extending the Concept of Component
      Interfaces: Experience with the Integrated Plasma Simulator*, in
      Component - Based High - Performance Computing (CBHPC) 2009, 2009,
      (extended abstract).
    - D Batchelor, G Alba, E D'Azevedo, G Bateman, DE Bernholdt, L Berry,
      P Bonoli, R Bramley, J Breslau, M Chance, J Chen, M Choi, W Elwasif,
      S Foley, G Fu, R Harvey, E Jaeger, S Jardin, T Jenkins, D Keyes, S
      Klasky, S Kruger, L Ku, V Lynch, D McCune, J Ramos, D Schissel, D
      Schnack, and J Wright, *Advances in Simulation of Wave Interactions with
      Extended MHD Phenomena*, in Horst Simon, editor, SciDAC 2009, 14-18 June
      2009, San Diego, California, USA, volume 180 of Journal of Physics:
      Conference Series, page 012054, Institute of Physics, 2009, 6pp.
    - Samantha S. Foley, Wael R. Elwasif, Aniruddha G. Shet, David E.
      Bernholdt, and Randall Bramley, *Incorporating Concurrent Component
      Execution in Loosely Coupled Integrated Fusion Plasma Simulation*, in
      Component-Based High-Performance Computing (CBHPC) 2008, 2008,
      (extended abstract).
    - D. Batchelor, C. Alba, G. Bateman, D. Bernholdt, L. Berry, P. Bonoli,
      R. Bramley, J. Breslau, M. Chance, J. Chen, M. Choi, W. Elwasif,
      G. Fu, R. Harvey, E. Jaeger, S. Jardin, T. Jenkins, D. Keyes,
      S. Klasky, S. Kruger, L. Ku, V. Lynch, D. McCune, J. Ramos, D. Schissel,
      D. Schnack, and J. Wright, *Simulation of Wave Interations with MHD*,
      in Rick Stevens, editor, SciDAC 2008, 14-17 July 2008, Washington, USA,
      volume 125 of Journal of Physics: Conference Series, page 012039,
      Institute of Physics, 2008.
    - Wael R. Elwasif, David E. Bernholdt, Lee A. Berry, and Don B.
      Batchelor, *Component Framework for Coupled Integrated Fusion
      Plasma Simulation*, in HPC-GECO/CompFrame 2007, 21-22 October,
      Montreal, Quebec, Canada, 2007.


   :Authors: Wael R. Elwasif, Samantha Foley, Aniruddha G. Shet
   :Organization: Center for Simulation of RF Wave Interactions
                  with Magnetohydrodynamics

"""
import sys
import argparse
import multiprocessing
import inspect
import socket
import logging
import os
import time
from ipsframework import platformspec
from ipsframework.messages import Message, ServiceRequestMessage, \
    ServiceResponseMessage, MethodInvokeMessage
from ipsframework.configurationManager import ConfigurationManager
from ipsframework.taskManager import TaskManager
from ipsframework.resourceManager import ResourceManager
from ipsframework.dataManager import DataManager
from ipsframework.componentRegistry import ComponentRegistry, ComponentID
from ipsframework.ipsExceptions import BlockedMessageException
from ipsframework.eventService import EventService
from ipsframework.cca_es_spec import initialize_event_service
from ipsframework.ips_es_spec import eventManager
from ipsframework._version import get_versions

if sys.version[0] != '3':  # noqa: E402
    print("IPS can is only compatible with Python 3.5 or higher")
    sys.exit(1)


class Framework:
    """Create an IPS Framework Instance to coordinate the execution of IPS simulations

    The Framework performs the following main tasks:

      * Initialize the different IPS managers that perform the bulk of the framework functionality
      * Manage communication queues, and route service requests from simulation
        components to appropriate managers.
      * Provide logging services to IPS managers.
      * Perfrom shutdown procedure on exit

    :param config_file_list: A list of simulation configuration files to be used
            in the simulaion. Each simulation configuration file must have the following
            parameters

            * *SIM_ROOT*    The root directory for the simulation
            * *SIM_NAME*    A name that identifies the simulation
            * *LOG_FILE*    The name of a log file that is used to capture logging and error information for this simulation.

            *SIM_ROOT*, *SIM_NAME*, and *LOG_FILE* must be unique across simulations.
    :type config_file_list: list

    :param log_file_name: A file name where Framework logging messages are placed.
    :type log_file_name: str

    :param platform_file_name: The name of the platform
            configuration file used in the simulation.  If not
            specified it will try to find the one installed in the
            share directory.
    :type platform_file_name: str

    :param debug: A flag indicating whether framework debugging messages are enabled (default = False)
    :type debug: bool

    :param verbose_debug: A flag adding more verbose framework debugging (default = False)
    :type verbose_debug: bool

    :param cmd_nodes: Computer nodes (default = 0)
    :type cmd_nodes: int

    :param cmd_ppn: Computer processor per nodes (default = 0)
    :type cmd_ppn: int
    """
    def __init__(self, config_file_list, log_file_name, platform_file_name=None,
                 debug=False, verbose_debug=False, cmd_nodes=0, cmd_ppn=0):
        # added compset_list for list of components to load config files for
        # command line option
        self.log_file_name = log_file_name
        if log_file_name == 'sys.stdout':
            self.log_file = sys.stdout
        else:
            self.log_file = open(os.path.abspath(log_file_name), 'w')
        # the multiprocessing queue
        self.in_queue = multiprocessing.Queue(0)
        # registry of components for calling
        self.comp_registry = ComponentRegistry()
        # reference to this class's component ID
        self.component_id = ComponentID(self.__class__.__name__, 'FRAMEWORK')
        # map of ports
        self.port_map = {}

        current_dir = inspect.getfile(inspect.currentframe())
        (self.platform_file_name, self.ipsShareDir) = \
            platformspec.get_share_and_platform(platform_file_name,
                                                current_dir)

        # config file list
        self.config_file_list = config_file_list

        # host is set in the configuration manager, not needed here???????
        self.host = socket.gethostname()
        self.logger = None
        self.service_handler = {}
        self.cur_time = time.time()
        self.start_time = self.cur_time
        self.event_service = EventService(self)
        initialize_event_service(self.event_service)
        self.event_manager = eventManager(self)
        self.config_manager = \
            ConfigurationManager(self, self.config_file_list, self.platform_file_name)
        self.resource_manager = ResourceManager(self)
        self.data_manager = DataManager(self)
        self.task_manager = TaskManager(self)
        # define a Handler which writes INFO messages or higher to the sys.stderr
        logger = logging.getLogger("FRAMEWORK")
        self.log_level = logging.WARNING
        if debug:
            self.log_level = logging.DEBUG
        # create handler and set level to debug
        logger.setLevel(self.log_level)
        ch = logging.StreamHandler(self.log_file)
        ch.setLevel(self.log_level)
        # create formatter
        formatter = logging.Formatter("%(asctime)s %(name)-15s %(levelname)-8s %(message)s")
        # add formatter to ch
        ch.setFormatter(formatter)
        # add ch to logger
        logger.addHandler(ch)
        self.logger = logger
        self.verbose_debug = verbose_debug
        self.outstanding_calls_list = []
        self.call_queue_map = {}

        # add the handler to the root logger
        try:
            # each manager should create their own event manager if they
            # want to send and receive events
            self.config_manager.initialize(self.data_manager,
                                           self.resource_manager,
                                           self.task_manager)
            self.task_manager.initialize(self.data_manager,
                                         self.resource_manager,
                                         self.config_manager)
            self.resource_manager.initialize(self.data_manager,
                                             self.task_manager,
                                             self.config_manager,
                                             cmd_nodes,
                                             cmd_ppn)
        except Exception:
            self.exception("Problem initializing managers")
            self.terminate_all_sims(status=Message.FAILURE)
            raise
        self.blocked_messages = []
        # SIMYAN: determine the sim_root for the Framework to use later
        fwk_comps = self.config_manager.get_framework_components()
        main_fwk_comp = self.comp_registry.getEntry(fwk_comps[0])
        self.sim_root = os.path.abspath(main_fwk_comp.services.get_config_param('SIM_ROOT'))

    def get_inq(self):
        """
        :return: handle to the Framework's input queue object
        :rtype: :class:`multiprocessing.Queue`
        """
        return self.in_queue

    def register_service_handler(self, service_list, handler):
        """
        Register a call back method to handle a list of framework service
        invocations.

        :param service_list: a list of service names to call *handler* when invoked by components.
             The service name must match the *target_method* parameter in :class:`messages.ServiceRequestMessage`.

        :param handler: a Python callable object that takes a :class:`messages.ServiceRequestMessage`.

        """
        for svc in service_list:
            self.service_handler[svc] = handler

    def _dispatch_service_request(self, msg):
        """
        Find and execute handler that corresponds to the *target_method* of
        *msg*, a :class:`messages.ServiceRequestMessage` object.  If handler
        not found an exception is raised.  The return value from the handler
        method is conveyed to the caller along the appropriate queue in a
        :class:`messages.ServiceResponseMessage`.  All exceptions are passed
        on to the caller, except for the
        :class:`ipsExceptions.BlockedMessageException`, which causes the
        message to be blocked until the request can be satisfied.
        """
        method_name = msg.target_method
        comp_id = msg.sender_id
        self.debug('Framework dispatching method: %s from %s', method_name, str(comp_id))
        try:
            handler = self.service_handler[method_name]
        except KeyError:
            self.exception("Unsupported method : %s", method_name)
            response_msg = ServiceResponseMessage(self.component_id,
                                                  comp_id,
                                                  msg.message_id,
                                                  Message.FAILURE,
                                                  Exception("Unsupported method : %s" % (method_name)))
        else:
            try:
                ret_val = handler(msg)
            except BlockedMessageException as e:
                if self.verbose_debug:
                    self.debug('Blocked message : %s', e.__str__())
                self.blocked_messages.append(msg)
                return
            except Exception as e:
                # self.exception('Exception handling service message: %s - %s', str(msg.__dict__), str(e))
                response_msg = ServiceResponseMessage(self.component_id,
                                                      comp_id,
                                                      msg.message_id,
                                                      Message.FAILURE, e)
            else:
                response_msg = ServiceResponseMessage(self.component_id,
                                                      comp_id,
                                                      msg.message_id,
                                                      Message.SUCCESS, ret_val)

        response_q = self.comp_registry.getComponentArtifact(comp_id,
                                                             'svc_response_q')
        response_q.put(response_msg)
        return

    def log(self, msg, *args):
        """
        Wrapper for :meth:`Framework.info`.
        """
        return self.info(msg, *args)

    def debug(self, msg, *args):
        """
        Produce **debugging** message in simulation log file. See :func:`logging.debug` for usage.
        """
        self.logger.debug(msg, *args)

    def info(self, msg, *args):
        """
        Produce **informational** message in simulation log file. See :func:`logging.info` for usage.
        """
        self.logger.info(msg, *args)

    def warning(self, msg, *args):
        """
        Produce **warning** message in simulation log file. See :func:`logging.warning` for usage.
        """
        self.logger.warning(msg, *args)

    def error(self, msg, *args):
        """
        Produce **error** message in simulation log file. See :func:`logging.error` for usage.
        """
        self.logger.error(msg, *args)

    def exception(self, msg, *args):
        """
        Produce **exception** message in simulation log file. See :func:`logging.exception` for usage.
        """
        self.logger.exception(msg, *args, exc_info=False)

    def critical(self, msg, *args):
        """
        Produce **critical** message in simulation log file. See :func:`logging.critical` for usage.
        """
        self.logger.critical(msg, *args)

    def _invoke_framework_comps(self, fwk_comps, method_name):
        """
        Invoke *method_name* on components in *fwk_comps* (list of component
        ids).  Typically, this is
        all framework-attached components (portal bridge).  The
        calling method blocks until all invocations terminate. No arguments
        are passed to the invoked methods.
        """

        outstanding_fwk_calls = []
        for comp_id in fwk_comps:
            msg = ServiceRequestMessage(self.component_id,
                                        self.component_id, comp_id,
                                        'init_call', method_name, 0)
            self.debug('Framework sending message %s ', msg.__dict__)
            call_id = self.task_manager.init_call(msg, manage_return=False)
            outstanding_fwk_calls.append(call_id)

        self.blocked_messages = []
        while len(outstanding_fwk_calls) > 0:
            self.debug("Framework waiting for message")
            msg = self.in_queue.get()
            self.debug("Framework received Message : %s", str(msg.__dict__))
            msg_list = [msg] + self.blocked_messages
            self.blocked_messages = []
            for msg in msg_list:
                self.debug('Framework processing message %s ', msg.message_id)
                if msg.__class__.__name__ == 'ServiceRequestMessage':
                    self._dispatch_service_request(msg)
                    continue
                elif msg.__class__.__name__ == 'MethodResultMessage':
                    self.debug('Received Result for call %s' %
                               (msg.call_id))
                    if msg.call_id not in outstanding_fwk_calls:
                        self.task_manager.return_call(msg)
                    else:
                        if msg.status == Message.FAILURE:
                            self.terminate_all_sims(status=Message.FAILURE)
                            raise msg.args[0]
                        outstanding_fwk_calls.remove(msg.call_id)
                else:
                    self.error('Framework received unexpected message : %s',
                               str(msg.__dict__))

    def run(self):
        """
        Run the communication outer loop of the framework.

        This method implements the core communication and message dispatch
        functionality of the framework. The main phases of execution for the
        framework are:

            1. Invoke the ``init`` method on all framework-attached components,
               blocking pending method call termination.
            2. Generate method invocation messages for the remaining public
               method in the framework-centric components (i.e. ``step`` and
               ``finalize``).
            3. Generate a queue of method invocation messages for all public
               framework accessible components in the simulations being run.
               framework-accessible components are made up of the **Init**
               component (if is exists), and the **Driver** component. The
               generated messages invoke the public methods ``init``,
               ``step``, and ``finalize``.
            4. Dispatch method invocations  for each framework-centric
               component and physics simulation in order.

        Exceptions that propagate to this method from the managed simulations
        causes the framework to abort any pending method invocation for the
        source simulation.
        Exceptions from framework-centeric component aborts further
        invocations to that component.

        When all method invocations have been dispatched (or aborted),
        :meth:`configurationManager.ConfigurationManager.terminate_sim` is called to trigger normal
        termination of all component processes.

        :return: Success status
        :rtype: bool
        """
        try:
            fwk_comps = self.config_manager.get_framework_components()
            sim_comps = self.config_manager.get_component_map()
            outstanding_sim_calls = {}
        except Exception:
            self.exception('encountered exception during fwk.run() initialization')
            self.terminate_all_sims(status=Message.FAILURE)
            return False

        # SIMYAN: get the runspaceInit_component and invoke its init() method
        # this creates the base directory and container file for the simulation
        # and copies the conf files into both and change directory to base dir
        main_fwk_comp = self.comp_registry.getEntry(fwk_comps[0])
        self.sim_root = os.path.abspath(main_fwk_comp.services.get_config_param('SIM_ROOT'))
        self._invoke_framework_comps(fwk_comps, 'init')

        try:
            # Each Framework Component is treated as a stand-alone simulation
            # generate the queues of invocation messages for each framework component
            for comp_id in fwk_comps:
                msg_list = []
                for method in ['step', 'finalize']:
                    req_msg = ServiceRequestMessage(self.component_id, self.component_id,
                                                    comp_id, 'init_call', method, 0)
                    msg_list.append(req_msg)

                outstanding_sim_calls[str(comp_id)] = msg_list

            # generate a queue of invocation messages for each simulation
            #   - list will look like: [init_comp.init(), init_comp.step(), init_comp.finalize(),
            #                           driver.init(), driver.step(), driver.finalize()]
            # these messages will be sent on a FIFO basis, thus running the init components,
            # then the corresponding drivers.
            for sim_name, comp_list in list(sim_comps.items()):
                msg_list = []
                self._send_monitor_event(sim_name, 'IPS_START', 'Starting IPS Simulation')
                self._send_dynamic_sim_event(sim_name=sim_name, event_type='IPS_START')
                comment = 'Nodes = %d   PPN = %d' % \
                          (self.resource_manager.num_nodes, self.resource_manager.ppn)
                self._send_monitor_event(sim_name, 'IPS_RESOURCE_ALLOC', comment)
                # SIMYAN: ordered list of methods to call
                methods = ['init', 'step', 'finalize']

                # SIMYAN: add each method call to the msg_list
                for comp_id in comp_list:
                    for method in methods:
                        req_msg = ServiceRequestMessage(self.component_id,
                                                        self.component_id,
                                                        comp_id,
                                                        'init_call', method, 0)
                        msg_list.append(req_msg)
                    # SIMYAN: add the msg_list to the outstanding sim calls
                    if msg_list:
                        outstanding_sim_calls[sim_name] = msg_list

        except Exception:
            self.exception('encountered exception during fwk.run() genration of call messages')
            self.terminate_all_sims(status=Message.FAILURE)
            return False

        # send off first round of invocations...
        try:
            for sim_name, msg_list in list(outstanding_sim_calls.items()):
                msg = msg_list.pop(0)
                self.debug('Framework sending message %s ', msg.__dict__)
                call_id = self.task_manager.init_call(msg, manage_return=False)
                self.call_queue_map[call_id] = msg_list
                self.outstanding_calls_list.append(call_id)
        except Exception:
            self.exception('encountered exception during fwk.run() sending first round of invocations (init of inits and fwk comps)')
            self.terminate_all_sims(status=Message.FAILURE)
            raise

        while len(self.outstanding_calls_list) > 0:
            if self.verbose_debug:
                self.debug("Framework waiting for message")
            # get new messages
            try:
                msg = self.in_queue.get()
            except Exception:
                continue
            if self.verbose_debug:
                self.debug("Framework received Message : %s", str(msg.__dict__))

            # add blocked messages to message list for reprocessing
            msg_list = [msg] + self.blocked_messages
            self.blocked_messages = []
            # process new and blocked messages
            for msg in msg_list:
                if self.verbose_debug:
                    self.debug('Framework processing message %s ', msg.message_id)

                sim_name = msg.sender_id.get_sim_name()
                if msg.__class__.__name__ == 'ServiceRequestMessage':
                    try:
                        self._dispatch_service_request(msg)
                    except Exception:
                        self.exception('Error dispatching service request message.')
                        self.terminate_all_sims(status=Message.FAILURE)
                        return False
                    continue
                elif msg.__class__.__name__ == 'MethodResultMessage':
                    if msg.call_id not in self.outstanding_calls_list:
                        self.task_manager.return_call(msg)
                        continue
                    # Message is a result from a framework invocation
                    self.outstanding_calls_list.remove(msg.call_id)
                    sim_msg_list = self.call_queue_map[msg.call_id]
                    del self.call_queue_map[msg.call_id]
                    if msg.status == Message.FAILURE:
                        self.error('received a failure message from component %s : %s',
                                   msg.sender_id, str(msg.args))
                        # No need to process remaining messages for this simulation
                        sim_msg_list = []
                        comment = 'Simulation Execution Error'
                        ok = False
                        # self.terminate_sim(status=Message.FAILURE)
                        # return False
                        self.send_terminate_msg(sim_name, Message.FAILURE)
                    else:
                        comment = 'Simulation Ended'
                        ok = True
                    try:
                        next_call_msg = sim_msg_list.pop(0)
                        call_id = self.task_manager.init_call(next_call_msg,
                                                              manage_return=False)
                        self.outstanding_calls_list.append(call_id)
                        self.call_queue_map[call_id] = sim_msg_list
                    except IndexError:
                        sim_comps = self.config_manager.get_component_map()  # Get any new dynamic simulations
                        if sim_name in list(sim_comps.keys()):
                            self._send_monitor_event(sim_name, 'IPS_END',
                                                     comment, ok)
                            self._send_dynamic_sim_event(sim_name, 'IPS_END', ok)
                            self.send_terminate_msg(sim_name, Message.SUCCESS)
                            self.config_manager.terminate_sim(sim_name)

        self.terminate_all_sims(Message.SUCCESS)
        self.event_service._print_stats()
        return True

    def _send_monitor_event(self, sim_name='', eventType='', comment='', ok='True'):
        """
        Publish a portal monitor event to the *_IPS_MONITOR* event topic.
        Event topics that start with an underscore are reserved for use by the
        IPS Framework and services.

          * *sim_name*: The name of the simulation to which this even belongs.
          * *eventType*: The type of the event.
          * *comment*: A string containing comment that describes the event.
          * *ok*: A string containing the values 'True' or 'False', based on
            whether the event indicates normal simulation execution, or an
            error condition.
        """
        if self.verbose_debug:
            self.debug('_send_monitor_event(%s - %s)', sim_name, eventType)
        portal_data = {}
        portal_data['code'] = 'Framework'
        # eventData['portal_runid'] = self.portalRunId
        portal_data['eventtype'] = eventType
        portal_data['ok'] = ok
        portal_data['comment'] = comment
        portal_data['walltime'] = '%.2f' % (time.time() - self.start_time)
        topic_name = '_IPS_MONITOR'
        # portal_data['phystimestamp'] = self.timeStamp
        get_config = self.config_manager.get_config_parameter
        if eventType == 'IPS_START':
            portal_data['state'] = 'Running'
            portal_data['host'] = get_config(sim_name, 'HOST')
            try:
                portal_data['outputprefix'] = get_config(sim_name, 'OUTPUT_PREFIX')
            except KeyError:
                pass
            try:
                portal_data['simname'] = sim_name
            except KeyError:
                pass
            try:
                portal_data['tag'] = get_config(sim_name, 'TAG')
            except KeyError:
                pass
            try:
                portal_data['user'] = get_config(sim_name, 'USER')
            except KeyError:
                pass
            try:
                portal_data['rcomment'] = get_config(sim_name, 'SIMULATION_DESCRIPTION')
            except KeyError:
                try:
                    portal_data['rcomment'] = get_config(sim_name, 'RUN_COMMENT')
                except KeyError:
                    portal_data['rcomment'] = 'SWIM simulation run'
            try:
                portal_data['tokamak'] = get_config(sim_name, 'TOKAMAK_ID')
            except KeyError:
                pass
            try:
                portal_data['shotno'] = get_config(sim_name, 'SHOT_NUMBER')
            except KeyError:
                pass
            try:
                portal_data['sim_runid'] = get_config(sim_name, 'RUN_ID')
            except KeyError:
                pass
            portal_data['startat'] = time.strftime('%Y-%m-%d|%H:%M:%S%Z',
                                                   time.localtime(self.start_time))
        elif eventType == 'IPS_END':
            portal_data['state'] = 'Completed'
            portal_data['stopat'] = time.strftime('%Y-%m-%d|%H:%M:%S%Z',
                                                  time.localtime())

        event_body = {}
        event_body['sim_name'] = sim_name
        event_body['sim_root'] = get_config(sim_name, 'SIM_ROOT')
        event_body['portal_data'] = portal_data

        if self.verbose_debug:
            self.debug('Publishing %s', str(event_body))
        self.event_manager.publish(topic_name, 'IPS_SIM', event_body)

    def _send_dynamic_sim_event(self, sim_name='', event_type='', ok=True):
        self.debug('_send_dynamic_sim_event(%s:%s)', event_type, sim_name)
        event_data = {}
        event_data['eventtype'] = event_type
        event_data['SIM_NAME'] = sim_name
        event_data['ok'] = ok
        topic_name = '_IPS_DYNAMIC_SIMULATION'
        self.debug('Publishing %s', str(event_data))
        self.event_manager.publish(topic_name, 'IPS_DYNAMIC_SIM', event_data)

    def send_terminate_msg(self, sim_name, status=Message.SUCCESS):
        """This method remotely invokes the method
        :meth:`component.Component.terminate` on all componnets in the
        IPS simulation ``sim_name``.

        :param sim_name: The simulation name from which all the components are terminated
        :type sim_name: str

        :param status: message status, defaults to :obj:`messages.Message.SUCCESS`
        :type status: :obj:`messages.Message.SUCCESS`, :obj:`messages.Message.FAILURE`

        """
        comp_ids = self.comp_registry.get_component_ids(sim_name)
        for comp_id in comp_ids:
            try:
                invocation_q = self.comp_registry.getComponentArtifact(comp_id,
                                                                       'invocation_q')
                call_id = self.task_manager.get_call_id()
                msg = MethodInvokeMessage(self.component_id, comp_id, call_id,
                                          'terminate', status)
                self.debug('Sending terminate message to %s', str(comp_id))
                invocation_q.put(msg)
            except Exception as e:
                self.exception('exception encountered while terminating comp %s', comp_id)
                print(e)

    def terminate_all_sims(self, status=Message.SUCCESS):
        """Terminate all active component instances.

        This method remotely invokes the method
        :meth:`component.Component.terminate` on all componnets in the
        IPS simulation.

        :param status: message status, defaults to :obj:`messages.Message.SUCCESS`
        :type status: :obj:`messages.Message.SUCCESS`, :obj:`messages.Message.FAILURE`

        """
        sim_names = self.config_manager.get_sim_names()
        for sim in sim_names:
            self.send_terminate_msg(sim, status)
        time.sleep(1)
        try:
            self.config_manager.terminate(status)
        except Exception:
            self.exception('exception encountered while cleaning up config_manager')
        # sys.exit(status)


def main(argv=None):
    """
    Check and parse args, create and run the framework.
    """
    platform_default = os.environ.get("IPS_PLATFORM_FILE")
    if platform_default:
        print("IPS using platform file :", platform_default)

    parser = argparse.ArgumentParser()
    parser.add_argument('--version', action='version', version="%(prog)s " + get_versions()['version'])
    parser.add_argument('--simulation', '-i', '--config', '-j',
                        required=True,
                        help='IPS simulation/config file')
    parser.add_argument('--platform', '-p', dest='platform_filename', default=platform_default,
                        required=not platform_default,
                        help='IPS platform configuration file')
    parser.add_argument('--debug', '-d', default=False, action='store_true',
                        help='Turn on debugging')
    parser.add_argument('--verbose', '-v', dest='verbose_debug', default=False, action='store_true',
                        help='Run IPS verbosely')
    parser.add_argument('--log', '-l', dest='log_file', default='sys.stdout',
                        help='IPS Log file')
    parser.add_argument('--nodes', '-n', dest='cmd_nodes', default='0',
                        type=int, help='Computer nodes')
    parser.add_argument('--ppn', '-o', dest='cmd_ppn', default='0',
                        type=int, help='Computer processor per nodes')

    options = parser.parse_args()

    cfgFile_list = options.simulation.split(',')

    try:
        fwk = Framework(cfgFile_list, options.log_file, options.platform_filename,
                        options.debug, options.verbose_debug,
                        options.cmd_nodes, options.cmd_ppn)
        fwk.run()
    except Exception:
        raise

    return 0


if __name__ == "__main__":
    print("Starting IPS")
    sys.stdout.flush()
    os.environ['IPS_INITIAL_CWD'] = os.getcwd()
    sys.exit(main())



# -------------------------------------------------------------------------------
# Copyright 2006-2020 UT-Battelle, LLC. See LICENSE for more information.
# -------------------------------------------------------------------------------
import sys


class Singleton:
    def __new__(cls, *param, **keywords):
        if '_the_instance' not in cls.__dict__:
            cls._the_instance = object.__new__(cls)
        return cls._the_instance


class ComponentID:
    """
    Object to facilitate the creation, serialization and deserialization of
    component ids.
    """
    delimiter = '@'
    seq_num = 0
    all_ids = {}

    @staticmethod
    def deserialize(comp_id_string):
        """
        Return the deserialized version of the component id.
        """
        tokens = comp_id_string.split(ComponentID.delimiter)
        if len(tokens) != 3:
            print('Invalid serialized component ID : ', comp_id_string)
            sys.exit(1)
        return ComponentID.all_ids[comp_id_string]

    def __init__(self, class_name, sim_name):
        """
        Create a component id from *class_name* and *sim_name*.
        """
        self.class_name = class_name
        self.sim_name = sim_name
        self.seq_num = ComponentID.seq_num
        ComponentID.seq_num += 1
        self.serialization = ''
        self.instance_name = ''
        self.serialization = self.get_serialization()
        self.instance_name = self.get_instance_name()
        ComponentID.all_ids[self.serialization] = self

    def __str__(self):
        """
        Return serialization.
        """
        return self.get_serialization()

    def __repr__(self):
        """
        Return string version (serialization).
        """
        return self.__str__()

    def __eq__(self, other):
        return str(self) == str(other)

    def get_instance_name(self):
        """
        Return instance name of component id.
        """
        if self.instance_name == '':
            self.instance_name = '@'.join([self.sim_name, self.class_name, str(self.seq_num)])
        return self.instance_name

    def get_serialization(self):
        """
        Return serialization.
        """
        if self.serialization == '':
            self.serialization = self.get_instance_name()
        return self.serialization

    def get_sim_name(self):
        """
        Return simulation name for the component.
        """
        return self.sim_name

    def get_class_name(self):
        """
        Return class name of component.
        """
        return self.class_name

    def get_seq_num(self):
        """
        Return sequence number of component.
        """
        return self.seq_num


class ComponentRegistry(Singleton):

    class RegistryEntry:
        """
        Container for queues and references associated with a component.
        """

        def __init__(self, svc_response_q, invocation_q, component_ref,
                     services, config):
            self.svc_response_q = svc_response_q
            self.invocation_q = invocation_q
            self.component_ref = component_ref
            self.services = services
            self.config = config

    def __init__(self):
        self.registry = {}

    def get_component_ids(self, sim_name):
        """
        Return all of the component ids associated with sim *sim_name*
        """
        ids = [ComponentID.deserialize(i) for i in self.registry
               if ComponentID.deserialize(i).get_sim_name() == sim_name]
        return ids

    def addEntry(self, component_id, svc_response_q, invocation_q,
                 component_ref, services, config):
        """
        Create a component registry entry for *component_id* and its
        associated queues, component ref, services and configuration
        information.
        """
        key = component_id.get_serialization()
        value = self.RegistryEntry(svc_response_q, invocation_q,
                                   component_ref,
                                   services, config)
        try:
            self.registry[key] = value
        except KeyError as e:
            print('Error creating component registery entry for ', key,
                  ' : ', str(e))
            raise e

    def removeEntry(self, component_id):
        key = component_id.get_serialization()
        try:
            del self.registry[key]
        except KeyError as e:
            print('Error removing component registry entry for ', key,
                  ' : ', str(e))
            raise

    # SIMYAN: this was added to provide an easy way to use the component
    # registry to get a registry entry
    def getEntry(self, component_id):
        """
        Return a registry entry.
        """
        key = component_id.get_serialization()
        try:
            entry = self.registry[key]
        except KeyError:
            print('No registry entry found for ', key)
            raise
        return entry

    def getComponentArtifact(self, component_id, artifact):
        """
        Return value of *artifact* in *component_id*'s registry entry.
        """
        key = component_id.get_serialization()
        try:
            entry = self.registry[key]
        except KeyError:
            print('No registry entry found for ', key)
            raise
        try:
            value = entry.__getattribute__(artifact)
        except KeyError:
            print('Invalid registry attribute : ', artifact)
            print('Possible values are : ', list(entry.__dict__.keys()))
            raise
        return value

    def setComponentArtifact(self, component_id, artifact, value):
        """
        Set the value of *artifact* in *component_id*'s registry entry to
        *value*.
        """
        key = component_id.get_serialization()
        try:
            entry = self.registry[key]
        except KeyError:
            print('No registry entry found for ', key)
            raise
        setattr(entry, artifact, value)



# -------------------------------------------------------------------------------
# Copyright 2006-2020 UT-Battelle, LLC. See LICENSE for more information.
# -------------------------------------------------------------------------------
import os
import sys
from .messages import Message


def get_share_and_platform(platform_file_name, ipsPathName):
    if platform_file_name:
        return platform_file_name, ''
    else:
        ipsPDir0 = os.path.dirname(ipsPathName)
        ipsPDir1 = os.path.dirname(ipsPDir0)
        ipsPDir2 = os.path.dirname(ipsPDir1)
        # This is if we've installed it
        pconf = os.path.join('share', 'platform.conf')
        if os.path.exists(os.path.join(ipsPDir1, pconf)):
            ipsShareDir = os.path.join(ipsPDir1, 'share')
        # This is looking in the build directory.
        elif os.path.exists(os.path.join(ipsPDir2, pconf)):
            ipsShareDir = os.path.join(ipsPDir2, 'share')
        else:
            print("Need to specify a platform file")
            sys.exit(Message.FAILURE)
        platform_file_name = os.path.join(ipsShareDir, 'platform.conf')
        return os.path.abspath(platform_file_name), ipsShareDir



#! /usr/bin/env python
# -------------------------------------------------------------------------------
# Copyright 2006-2020 UT-Battelle, LLC. See LICENSE for more information.
# -------------------------------------------------------------------------------

import os
import sys
import getopt
import inspect
import subprocess
import tempfile
import re
import time
from multiprocessing.connection import Client
from ipsframework.configobj import ConfigObj
from ipsframework import platformspec
from ipsframework.ipsutil import which


class DakotaDynamic:
    def __init__(self, dakota_cfg, log_file, platform_filename, debug, ips_config_template, restart_file):
        self.dakota_cfg = dakota_cfg
        self.log_file = log_file
        self.platform_fname = platform_filename
        self.debug = debug
        self.config_template = ips_config_template
        self.template_conf = None
        self.platform_conf = None
        self.dakota_conf = None
        self.master_conf = ConfigObj()
        self.restart_file = restart_file

    def run(self):  # noqa: C901
        alt_paths = []

        """
        Dakota Configuration
        Control variables expected in the format COMPONENT__VARIABLE (two _)
        """
        try:
            self.dakota_conf = [t.strip() for t in open(self.dakota_cfg).readlines()]
        except Exception:
            raise

        """
        Platform Configuration
        """
        # parse file
        try:
            current_dir = inspect.getfile(inspect.currentframe())
            (self.platform_fname, self.ipsShareDir) = \
                platformspec.get_share_and_platform(self.platform_fname,
                                                    current_dir)

            if self.ipsShareDir:
                haveComp = False
                if os.path.exists(os.path.join(self.ipsShareDir, 'component-generic.conf')):
                    comp_conf_file = os.path.join(self.ipsShareDir, 'component-generic.conf')
                    comp_confgobj = ConfigObj(comp_conf_file, interpolation='template',
                                              file_error=True)
                    haveComp = True

                self.platform_conf = ConfigObj(self.platform_fname, interpolation='template',
                                               file_error=True)
                if haveComp:
                    self.platform_conf.merge(comp_confgobj)
            else:
                self.platform_conf = ConfigObj(self.platform_fname, interpolation='template',
                                               file_error=True)

        except (IOError, SyntaxError):
            raise

        """
        Master Config file
        """
        # parse file
        try:
            self.template_conf = ConfigObj(self.config_template, interpolation='template', file_error=True)
        except (IOError, SyntaxError):
            raise
        for k in list(self.platform_conf.keys()):
            if k not in list(self.template_conf.keys()):
                self.template_conf[k] = self.platform_conf[k]

        # Import environment variables into config file
        # giving precedence to config file definitions in case of duplicates
        for (k, v) in os.environ.items():
            if k not in list(self.template_conf.keys()) and not any(x in v for x in '{}()$'):
                self.template_conf[k] = v

        alt_paths.append(self.template_conf['IPS_ROOT'])
        alt_paths.append(os.path.join(self.template_conf['IPS_ROOT'], 'bin'))
        alt_paths.append(os.path.join(self.template_conf['IPS_ROOT'], 'framework/src'))

        new_dakota_config = self.dakota_cfg + '.resolved'
        comp_vars = {}
        for line in self.dakota_conf:
            if not line:
                continue
            if line[0] == '#':
                continue
            tokens = line.split()
            if tokens[0] == 'descriptors':
                for token in tokens[1:]:
                    raw_token = token.replace("'", '').replace('"', '')
                    try:
                        (comp, var) = raw_token.split('__')
                    except ValueError:
                        print('Error: variable %s not of the form COMP__VARNAME')
                        raise
                    comp_vars[comp] = var
            elif tokens[0] == 'analysis_driver':
                raw_prog = line.split('=')[1]
                prog = raw_prog.strip(' "\'')
                exec_prog = which(prog, alt_paths)
                if not exec_prog:
                    raise Exception('Error: analysis driver %s not found in path' % prog)
                line.replace(prog, exec_prog)
            elif tokens[0] == 'system':
                if 'asynchronous' not in line:
                    raise Exception('Asynchronous specification missing from DAKOTA system line in interface section')
                match = re.search(r'evaluation_concurrency\s*=\s*\d*', line)
                if match:
                    conc_tokens = match.group(0).split(' =')
                    self.batch_size = int(conc_tokens[1])
                    print('Using evaluation_concurrency = ', self.batch_size)
                else:
                    print('Missing evaluation_concurrency spec, using default value of %d' % (self.batch_size))

        self.master_conf['PORTS'] = {'NAMES': 'DRIVER'}
        self.master_conf['PORTS']['DRIVER'] = {'IMPLEMENTATION': 'DAKOTA_BRIDGE'}
        self.master_conf['PORTS']['INIT'] = {'IMPLEMENTATION': ''}
        driver_conf = {}
        driver_conf['CLASS'] = 'DAKOTA'
        driver_conf['SUB_CLASS'] = 'BRIDGE'
        driver_conf['NAME'] = 'Driver'
        driver_conf['NPROC'] = 1
        driver_conf['BIN_PATH'] = os.path.join(self.template_conf['IPS_ROOT'], 'bin')
        driver_conf['BIN_DIR'] = driver_conf['BIN_PATH']
        driver_conf['INPUT_DIR'] = '/dev/null'
        driver_conf['INPUT_FILES'] = ''
        driver_conf['OUTPUT_FILES'] = ''
        script = os.path.join(self.template_conf['IPS_ROOT'],
                              'bin', 'dakota_bridge.py')
        if os.path.isfile(script):
            driver_conf['SCRIPT'] = script
        else:
            script = os.path.join(self.template_conf['IPS_ROOT'], 'framework',
                                  'src', 'dakota_bridge.py')
            if os.path.isfile(script):
                driver_conf['SCRIPT'] = script
            else:
                raise Exception('Error: unable to locate dakota_bridge.py in \
IPS_ROOT/bin or IPS_ROOT/framework/src')
        self.master_conf['DAKOTA_BRIDGE'] = driver_conf

        for comp in comp_vars:
            if comp == '':
                continue
            try:
                self.template_conf[comp]
            except KeyError:
                print('Error: missing component %s in IPS configuration file')
                raise

        sim_root = self.template_conf['SIM_ROOT']
        try:
            os.makedirs(sim_root)
        except OSError as oserr:
            if oserr.errno != 17:
                print('Error creating Simulation directory %s : %d %s' %
                      (sim_root, oserr.errno, oserr.strerror))
                raise

        for (k, v) in self.template_conf.items():
            if k not in list(self.master_conf.keys()):
                try:
                    list(v.keys())
                except Exception:
                    self.master_conf[k] = v

        self.master_conf.filename = os.path.join(sim_root, 'dakota_bridge_%d.conf' % (os.getpid()))
        self.master_conf.write()

        sock_address = os.path.join(tempfile.gettempdir(), 'ips_dynamic_%d.tmp' % (os.getpid()))
        os.environ['IPS_DAKOTA_platform'] = os.path.abspath(self.platform_fname)
        os.environ['IPS_DAKOTA_config'] = os.path.abspath(self.config_template)
        os.environ['IPS_DAKOTA_runid'] = str(os.getpid())
        os.environ['IPS_DAKOTA_SOCKET_ADDRESS'] = sock_address

        with open(new_dakota_config, 'w') as fd:
            for line in self.dakota_conf:
                fd.write('%s\n' % (line))

        ips = which('ips.py', alt_paths)
        if not ips:
            raise Exception('Error: ips not found in path.')

        if self.restart_file:
            if not os.path.isfile(self.restart_file):
                raise Exception("Error accessing DAKOTA restart file %s" % (self.restart_file))

        cmd = '%s --all --simulation=%s --platform=%s --verbose' % (ips, self.master_conf.filename,
                                                                    os.environ['IPS_DAKOTA_platform'])
        if self.log_file:
            cmd += ' --log=' + self.log_file

        if self.debug:
            cmd += '  --debug'

        print('cmd =', cmd)
        ips_server_proc = subprocess.Popen(cmd, shell=True)
        print('%s  Launched IPS' % (time.strftime("%b %d %Y %H:%M:%S", time.localtime())))
        sys.stdout.flush()
        msg = {'SIMSTATUS': 'START'}
        num_trials = 30
        for trials in range(num_trials):
            try:
                conn = Client(sock_address, 'AF_UNIX')
                conn.send(msg)
                response = conn.recv()
            except Exception as inst:
                print('%s  %d ips_dakota_dynamic connecting to IPS dakota bridge' %
                      (time.strftime("%b %d %Y %H:%M:%S", time.localtime()),
                       trials), type(inst), str(inst))
                sys.stdout.flush()
                if trials == num_trials - 1:
                    ips_server_proc.kill()
                    raise
                else:
                    time.sleep(5)
            else:
                print('%s  ips_dakota_dynamic received response from IPS ' %
                      (time.strftime("%b %d %Y %H:%M:%S", time.localtime())), str(response))
                conn.close()
                break

        if self.restart_file:
            command = "dakota -read_restart %s -input %s" % (self.restart_file, new_dakota_config)
        else:
            command = 'dakota %s ' % new_dakota_config
        dakota_logfile = open('dakota_%s.log' % (str(os.getpid())), 'w')
        proc = subprocess.Popen(command, shell=True, stdout=dakota_logfile, stderr=subprocess.STDOUT)
        print('%s  Launched DAKOTA' % (time.strftime("%b %d %Y %H:%M:%S", time.localtime())))
        sys.stdout.flush()
        proc.wait()

        msg = {'SIMSTATUS': 'END'}
        num_trials = 1
        for trials in range(num_trials):
            try:
                conn = Client(sock_address, 'AF_UNIX')
                conn.send(msg)
                response = conn.recv()
            except Exception as inst:
                print('%s  %d ips_dakota_dynamic connecting to IPS dakota bridge' %
                      (time.strftime("%b %d %Y %H:%M:%S", time.localtime()), trials), type(inst), str(inst))
                sys.stdout.flush()
                if trials == num_trials - 1:
                    ips_server_proc.kill()
                    raise
                else:
                    time.sleep(5)
            else:
                print('%s  ips_dakota_dynamic received response from IPS ' %
                      (time.strftime("%b %d %Y %H:%M:%S", time.localtime())), str(response))
                conn.close()
                break

        ips_server_proc.wait()


def printUsageMessage():
    print("Usage: ips_dakota_dynamic --dakotaconfig=DAKOTA_CONFIG_FILE --simulation=CONFIG_FILE_NAME "
          "--platform=PLATFORM_FILE_NAME --log=LOG_FILE_NAME --restart=DAKOTA_RESTART_FILE [--debug]")


def main(argv=None):

    ips_config_file = None
    platform_filename = ''
    # parse command line arguments
    if argv is None:
        argv = sys.argv
        first_arg = 1
    else:
        first_arg = 0

    try:
        opts, _ = getopt.gnu_getopt(argv[first_arg:], '',
                                    ["dakotaconfig=", "simulation=", "platform=", "log=", "restart=", "debug"])
    except getopt.error as msg:
        print('Invalid command line arguments', msg)
        printUsageMessage()
        return 1
    debug = False
    log_file_name = None
    dakota_cfg = None
    platform_filename = None
    restart_file = None
    for arg, value in opts:
        if arg == '--simulation':
            ips_config_file = value
        elif arg == '--log':
            log_file_name = value
        elif arg == '--platform':
            platform_filename = value
        elif arg == '--dakotaconfig':
            dakota_cfg = value
        elif arg == '--restart':
            restart_file = value
        elif arg == '--debug':
            debug = True

    if (not ips_config_file or not dakota_cfg):
        printUsageMessage()
        return 1
    try:
        sweep = DakotaDynamic(dakota_cfg, log_file_name, platform_filename, debug, ips_config_file, restart_file)
        sweep.run()
    except Exception:
        raise
    return 0

# ----- end main -----


if __name__ == "__main__":
    sys.stdout.flush()
    sys.exit(main())



# -------------------------------------------------------------------------------
# Copyright 2006-2020 UT-Battelle, LLC. See LICENSE for more information.
# -------------------------------------------------------------------------------
import os
import shutil
import time
import glob
try:
    import Pyro4
except ImportError:
    pass

remote_copy_fun = None


def which(program, alt_paths=None):
    def is_exe(fpath):
        return os.path.exists(fpath) and os.access(fpath, os.X_OK)

    fpath, _ = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

        # Trust locations in platform file over those in environment path
        if alt_paths:
            for path in alt_paths:
                exe_file = os.path.join(path, program)
                if is_exe(exe_file):
                    return exe_file

    return None


def copyFiles(src_dir, src_file_list, target_dir, prefix='', keep_old=False):
    """
       Copy files in *src_file_list* from *src_dir* to *target_dir* with an
       optional prefix.  If *keep_old* is ``True``, existing files in
       *target_dir* will not be overridden, otherwise files can be clobbered
       (default).
       Wild-cards in file name specification are allowed.
    """

    global remote_copy_fun
    use_data_server = os.getenv('USE_DATA_SERVER', "DATA_SERVER_NOT_USED")
    if use_data_server != "DATA_SERVER_NOT_USED":
        if not remote_copy_fun:
            data_server = Pyro4.Proxy("PYRONAME:DataServer")
        data_server.copyFiles(src_dir, src_file_list, target_dir, prefix, keep_old)
        return

    try:
        file_list = src_file_list.split()
    except AttributeError:  # srcFileList is not a string
        file_list = src_file_list

    globbed_file_list = []
    for src_file in file_list:

        if not target_dir == src_dir:
            src_file_full = os.path.join(src_dir, src_file)

            if os.path.isfile(src_file_full):
                globbed_file_list += [src_file_full]
            else:
                globbed_files = glob.glob(src_file_full)
                if len(globbed_files) > 0:
                    globbed_file_list += globbed_files
                else:
                    raise Exception('No such file : %s' % (src_file_full))

    # ------------------------------------------------------------------#
    #  for each file in globbed_file_list, copy it from src_dir to target_dir #
    # ------------------------------------------------------------------#
    for src_file in globbed_file_list:
        target = prefix + os.path.basename(src_file)
        target_file = os.path.join(target_dir, target)
        if (os.path.isfile(target_file) and os.path.samefile(src_file, target_file)):
            continue
    # Do not overwrite existing target files.
        if (keep_old and os.path.isfile(target_file)):
            for i in range(1000):
                new_name = target_file + '.' + str(i)
                if os.path.isfile(new_name):
                    continue
                target_file = new_name
                break

        (head, _) = os.path.split(os.path.abspath(target_file))
        try:
            os.makedirs(head)
        except OSError as oserr:
            if oserr.errno != 17:
                print('Error creating directory %s : %s' % (head, oserr.strerror))
                raise
        try:
            shutil.copy(src_file, target_file)
        except Exception:
            raise


def getTimeString(timeArg=None):
    """
    Return a string representation of *timeArg*. *timeArg* is expected
    to be an appropriate object to be processed by :py:meth:`time.strftime`.
    If *timeArg* is ``None``, current time is used.
    """
    if timeArg is None:
        arg = time.localtime()
    else:
        arg = timeArg
    return time.strftime('%Y-%m-%d|%H:%M:%S%Z', arg)



# -------------------------------------------------------------------------------
# Copyright 2006-2020 UT-Battelle, LLC. See LICENSE for more information.
# -------------------------------------------------------------------------------
"""
This file hosts the TopicManager class that manages the set of events and
listeners for a given topic. For reasons of efficient storage and O(1)-time
processing, TopicManager maintains a single list of events for the topic, and
marks the start of next event delivery in this list for individual listeners.
Listeners receive only the events that are published _after_ they have
registered. This simplifies event cleanup to a great deal and also avoids the
uncertainty that arises when trying to deliver prior events. Events not pending
processing by any listener are purged periodically as a result of a listener
activity like processing or unregistering. For cases where the event list could
grow unbounded in the absence of any listener activity for prolonged periods of
time, we could define a 'limitPendingEvents' parameter to denote a bound on the
count of pending events exceeding which triggers an event cleanup to remove
events that outlive a 'timeToLive' parameter.
"""

from .cca_es_spec import EventServiceException, Event
from .debug import debug


class TopicManager:
    def __init__(self, limitPendingEvents=10):
        """ eventList is the common listing of events posted to a topic. """
        self.eventList = []

        """
        listenerDirectory holds listeners alongside their event list
        markers.
        """
        self.listenerDirectory = {}

        """
        maxPendingEvents denotes the maximum number of events pending for this
        topic over the lifetime of an actual run.
        """
        self.maxPendingEvents = 0

        """
        A parameter to bound the memory for individual event queues; specifies an
        upper limit on the number of events permitted to be pending for this topic
        at any point in time. Not pressed into service yet.
        """
        self.limitPendingEvents = limitPendingEvents

        debug.output("TopicManager.__init__")
        self.printEventsAndListeners()

    """
    Append the event to the event list.
    """

    def sendEvent(self, theEvent):
        """
        A new event is appended to the end of the event list provided there is at
        least one registered listener. This is in accordance with the policy that
        listeners receive only those events that are published _after_ they
        have registered. An event that is published before any listener has
        registered will then never be consumed and hence is not stored. This also
        prevents the event list from growing in an unbounded manner when there are
        no registered listeners.
        """
        if len(self.listenerDirectory) > 0:
            self.eventList.append(theEvent)
            eventList_len = len(self.eventList)
            if eventList_len > self.maxPendingEvents:
                self.maxPendingEvents = eventList_len
        debug.output("TopicManager.sendEvent")
        self.printEventsAndListeners()

    def registerListener(self, listenerid):
        """
        For a new listener, the event list marker is initialized to the end of the
        event list in accordance with the policy that listeners receive only those
        events that are published _after_ they have registered.
        """
        if listenerid not in self.listenerDirectory:
            self.listenerDirectory[listenerid] = len(self.eventList)
            debug.output("TopicManager.registerListener")
            self.printEventsAndListeners()
        else:
            raise EventServiceException("Event listener registered earlier.")

    """
    A listener activity like unregistering or processing triggers cleanup of
    events having no pending listeners, followed by resetting of list markers
    of all registered listeners.
    """

    def cleanupEvents(self, listenerid):
        self.listenerDirectory[listenerid] = len(self.eventList)

        """ First determine the oldest pending event. """
        oldestPendingEvent = min(self.listenerDirectory.values())

        """
        Reset current listeners' list markers and remove events having no
        pending listeners.
        """
        if oldestPendingEvent > 0:
            del self.eventList[:oldestPendingEvent]
            for listener_id in self.listenerDirectory:
                self.listenerDirectory[listener_id] -= oldestPendingEvent

    """
    A listener is unregistered by first performing an event cleanup, followed by
    deletion of the listener from listenerDirectory.
    """

    def unregisterListener(self, listenerid):
        self.cleanupEvents(listenerid)
        del self.listenerDirectory[listenerid]
        debug.output("TopicManager.unregisterListener")
        self.printEventsAndListeners()

    """
    Returns events posted since the last fetch for a listener.
    """

    def getEventListForListener(self, listenerid):
        eventListForListener = []
        for theEvent in self.eventList[self.listenerDirectory[listenerid]:]:
            eventListForListener.append(Event(theEvent.header, theEvent.body))
        self.cleanupEvents(listenerid)
        debug.output("TopicManager.getEventListForListener")
        self.printEventsAndListeners()
        return eventListForListener

    """
    Print out the contents for debugging.
    """

    def printEventsAndListeners(self):
        string = ":::::::::\n" + "List of events:"
        i = 0
        for i in range(len(self.eventList)):
            string += "\n" + str(i) + "---" + str(self.eventList[i])
        string += "\n\n" + "List of listeners:"
        debug.output(string)
        sortedKeys = sorted(self.listenerDirectory.keys())
        for listenerid in sortedKeys:
            string = "event = " + str(self.listenerDirectory[listenerid])
            debug.output(string, listenerid)
        debug.output(":::::::::")

    """
    Gives a profile of events posted to this topic, currently just
    maxPendingEvents.
    """

    def getEventStats(self):
        return self.maxPendingEvents


if __name__ == "__main__":
    em = TopicManager()

    em.registerListener(1)
    em.registerListener(2)

    print("-----")
    el = em.getEventListForListener(1)
    for event in el:
        print("1:" + str(event))
    print("-----\n")

    h0 = {}
    h0[0] = 0
    b0 = {}
    b0["zero"] = "zero"
    em.sendEvent(Event(h0, b0))

    h1 = {}
    h1[1] = 1
    b1 = {}
    b1["one"] = "one"
    em.sendEvent(Event(h1, b1))

    print("-----")
    el = em.getEventListForListener(1)
    for event in el:
        print("1:" + str(event))
    print("-----\n")

    em.unregisterListener(1)

    print("-----")
    el = em.getEventListForListener(2)
    for event in el:
        print("2:" + str(event))
    print("-----\n")

    em.unregisterListener(2)



# -------------------------------------------------------------------------------
# Copyright 2006-2020 UT-Battelle, LLC. See LICENSE for more information.
# -------------------------------------------------------------------------------
"""
Node structures for RM are implemented here for convenience.
"""

# local version


class Node:
    """
    Models a node in the allocation.

      * *name*: name of node, typically actual name from resource detection phase.
      * *task_ids*, *owners*: identifiers for the tasks and components that are currently using the node.
      * *allocated*, *available*: list of sockets that have cores allocated and available.  A socket may appear in both lists if it is only partially allocated.
      * *sockets*: list of sockets belonging to this node
      * *avail_cores*: number of cores that are currently available.
      * *total_cores*: total number of cores that can be allocated on this node.
      * *status*: indicates if the node is 'UP' or 'DOWN'.  Currently not used, all nodes are considered functional..
    """

    def __init__(self, name, socks, cores, p):
        self.status = 'UP'
        self.name = name
        self.task_ids = []
        self.owners = []
        self.allocated = []  # sockets allocated
        self.available = []  # sockets available
        self.sockets = []
        if isinstance(p, int):
            self.avail_cores = p
        else:
            self.avail_cores = len(p)
        cps = cores // socks

        self.total_cores = self.avail_cores
        c = 0
        s = 0
        i = 0
        while c < self.total_cores:
            if c + cps <= self.total_cores:
                if isinstance(p, int):
                    self.sockets.append(Socket(s, cps))
                else:
                    self.sockets.append(Socket(s, cps, p[i:cps]))
                    i += cps
                self.available.append(s)
                s += 1
                c += cps
            else:
                if isinstance(p, int):
                    self.sockets.append(Socket(s, p - c))
                    c = p
                else:
                    self.sockets.append(Socket(s, len(p) - c, p[i:]))
                    i = len(p)
                    c = len(p)
                self.available.append(s)

    def print_sockets(self, fname=''):
        """
        Pretty print of state of sockets.
        """
        if fname:
            for sock in self.sockets:
                print("    socket:", sock.name, file=fname)
                print("    availablilty:", sock.avail_cores, file=fname)
                print("    task ids:", sock.task_ids, file=fname)
                print("    owners:", sock.owners, file=fname)
                print("    cores:", sock.total_cores, file=fname)
                sock.print_cores(fname)

        else:
            for sock in self.sockets:
                print("    socket:", sock.name)
                print("    availablilty:", sock.avail_cores)
                print("    task ids:", sock.task_ids)
                print("    owners:", sock.owners)
                print("    cores:", sock.total_cores)
                sock.print_cores()

    def allocate(self, whole_nodes, whole_sockets, tid, o, procs):
        """
        Mark *procs* number of cores as allocated subject to the values of
        *whole_nodes* and *whole_sockets*.  Return the number of cores
        allocated and their corresponding slots, a list of strings of the form:

          <socket name>:<core name>
        """
        # add tid and o to lists
        slots = []
        self.task_ids.append(tid)
        self.owners.append(o)
        k = 0   # number of cores allocated

        if whole_nodes:
            for sock in self.sockets:
                slots.extend(sock.allocate(whole_sockets, tid, o,
                                           sock.avail_cores))
                self.available.remove(sock.name)
                self.allocated.append(sock.name)
                k += sock.total_cores
        elif whole_sockets:
            for sock in self.sockets:
                if sock.avail_cores == sock.total_cores:
                    slots.extend(sock.allocate(whole_sockets, tid, o,
                                               sock.avail_cores))
                    self.available.remove(sock.name)
                    self.allocated.append(sock.name)
                    k += sock.total_cores
                    if k >= procs:
                        break
        else:
            for sock in self.sockets:
                if sock.avail_cores > procs - k:
                    slots.extend(sock.allocate(whole_sockets, tid, o,
                                               procs - k))
                    k = procs
                    if sock.name not in self.allocated:
                        self.allocated.append(sock.name)
                    if sock.avail_cores == 0:
                        self.available.remove(sock.name)

                elif sock.avail_cores > 0:  # sock.avail_cores < procs - k
                    k += sock.avail_cores
                    slots.extend(sock.allocate(whole_sockets, tid, o,
                                               sock.avail_cores))
                    if sock.name not in self.allocated:
                        self.allocated.append(sock.name)
                    if sock.avail_cores == 0:
                        self.available.remove(sock.name)
                if k >= procs:
                    break

        self.avail_cores -= k
        return k, slots

    def release(self, tid, o):
        """
        Mark cores used by task *tid* and component *o* as available.  Return
        the number of cores released.
        """
        # remove tid and o from lists
        self.task_ids.remove(tid)
        self.owners.remove(o)
        ac = 0
        start = self.avail_cores
        for s in self.sockets:
            if tid in s.task_ids:
                s.release(tid)
                if s.name not in self.available:
                    self.available.append(s.name)
                elif s.avail_cores == s.total_cores:
                    self.allocated.remove(s.name)
            ac += s.avail_cores

        # update avail_cores
        self.avail_cores = ac
        return self.avail_cores - start


class Socket:
    """
    Models a socket in a node.

      * *name*: identifier for the socket
      * *task_ids*, *owners*: identifiers for the tasks and components that
        are currently using the socket.
      * *allocated*, *available*: lists of cores that are allocated
        and available.
      * *cores*: list of :py:class:`.Core` objects belonging to this socket
      * *avail_cores*: number of cores that are currently available.
      * *total_cores*: total number of cores that can be allocated on this
        socket.
    """

    def __init__(self, name, cps, coreids=[]):
        """
        s = number of sockets (per node)
        c = number of cores (per node)
        """
        self.name = name
        self.task_ids = []
        self.owners = []
        self.allocated = []
        self.available = []
        self.cores = []
        self.avail_cores = cps
        self.total_cores = cps
        self.my_tasks = {}  # tid: (owner, cores, num_procs)
        if coreids:
            for c in coreids:
                self.cores.append(Core(c))
                self.available.append(c)
        else:
            for c in range(cps):
                self.cores.append(Core(c))
                self.available.append(c)

    def print_cores(self, fname=''):
        """
        Pretty print of state of cores.
        """
        if fname:
            for c in self.cores:
                print("      core:", c.name, end=' ', file=fname)
                if c.is_available:
                    print(" - available", file=fname)
                else:
                    print(" - task_id:", c.task_id, end=' ', file=fname)
                    print(" - owner:", c.owner, file=fname)
        else:
            for c in self.cores:
                print("      core:", c.name, end=' ')
                if c.is_available:
                    print(" - available")
                else:
                    print(" - task_id:", c.task_id, end=' ')
                    print(" - owner:", c.owner)

    def allocate(self, whole, tid, o, num_procs):
        """
        Mark *num_procs* cores as allocated subject to the value of *whole*.
        Return a list of strings of the form:

          <socket name>:<core name>
        """
        self.task_ids.append(tid)
        self.owners.append(o)
        slots = []
        k = 0

        if whole:
            # fill the whole socket!
            for c1 in self.cores:
                self.available.remove(c1.name)
                slots.append(str(self.name) + ":" +
                             str(c1.allocate(tid, o)))
                self.allocated.append(c1.name)
                k += 1
        else:
            for c1 in self.cores:
                if c1.is_available:
                    self.available.remove(c1.name)
                    slots.append(str(self.name) + ":" +
                                 str(c1.allocate(tid, o)))
                    self.allocated.append(c1.name)
                    k += 1
                if k == num_procs:
                    break
        self.avail_cores -= k
        self.my_tasks.update({tid: (o, k, num_procs)})
        return slots

    def release(self, tid):
        """
        Mark cores that are allocated to task *tid* as available.  Return
        number of cores set to available.
        """
        o, k, _ = self.my_tasks[tid]
        self.task_ids.remove(tid)
        self.owners.remove(o)  # make sure it just removes one instance
        count = 0

        # release cores
        for c in self.cores:
            if c.task_id == tid:
                self.allocated.remove(c.name)
                c.release()
                self.available.append(c.name)
                count += 1
        if count != k:
            print("<<<error>>>")
        # set avail_cores
        self.avail_cores += k
        return count


class Core:
    """
    Models a core of a socket.

      * *name*: name of core
      * *is_available*: boolean value indicating the availability of the core.
      * *task_id*, *owner*: identifiers of the task and component using the core.
    """

    def __init__(self, name):
        self.name = name
        self.is_available = True
        self.task_id = -1
        self.owner = ''

    def allocate(self, tid, o):
        """
        Mark core as allocated.
        """
        if self.is_available:
            self.is_available = False
            self.task_id = tid
            self.owner = o
            return self.name
        else:
            print("trying to allocate core that is not available")
            raise RuntimeError("trying to allocate core that is not available")

    def release(self):
        """
        Mark core as available.
        """
        if self.is_available:
            print("warning: trying to release core when not in use")
        else:
            self.is_available = True
            self.task_id = -1
            self.owner = ''



# -------------------------------------------------------------------------------
# Copyright 2006-2020 UT-Battelle, LLC. See LICENSE for more information.
# -------------------------------------------------------------------------------


class EventServiceProxy:
    def getTopic(self, topicName):
        pass

    def existsTopic(self, topicName):
        pass

    def registerSubscriber(self):
        pass

    def unregisterSubscriber(self, subscriberid):
        pass

    def getSubscription(self, subscriberid, subscriptionName):
        pass

    def processEvents(self, subscriberid):
        pass

    def sendEvent(self, topicName, eventName, eventBody):
        pass

    def createListener(self):
        pass

    def registerEventListener(self, subscriberid, subscriptionName, listenerKey, listenerid, refListener):
        pass

    def unregisterEventListener(self, subscriberid, subscriptionName, listenerKey):
        pass

    def removeSubscription(self, subscriberid, subscriptionName):
        pass


# TODO: Is eventService.py the right placeholder for this class?#
class EventServiceFwkProxy(EventServiceProxy):
    def __init__(self, event_service):
        self.event_service = event_service
        self.listenerDirectory = {}

    def getTopic(self, topicName):
        return self.event_service.getTopic(topicName)

    def existsTopic(self, topicName):
        return self.event_service.existsTopic(topicName)

    def registerSubscriber(self):
        return self.event_service.registerSubscriber()

    def unregisterSubscriber(self, subscriberid):
        listenerList = self.event_service.unregisterSubscriber(subscriberid)
        for listenerid in listenerList:
            self._removeEventListener(listenerid)

    def getSubscription(self, subscriberid, subscriptionName):
        self.event_service.getSubscription(subscriberid, subscriptionName)

    def processEvents(self, subscriberid):
        eventList = self.event_service.processEvents(subscriberid)
        for listenerid in list(eventList.keys()):
            for topicName in list(eventList[listenerid].keys()):
                for theEvent in eventList[listenerid][topicName]:
                    self.listenerDirectory[listenerid].processEvent(topicName, theEvent)

    def sendEvent(self, topicName, eventName, eventBody):
        self.event_service.sendEvent(topicName, eventName, eventBody)

    def createListener(self):
        return self.event_service.createListener()

    def registerEventListener(self, subscriberid, subscriptionName, listenerKey, listenerid, refListener):
        self.event_service.registerEventListener(subscriberid, subscriptionName, listenerKey, listenerid)
        self._addEventListener(listenerid, refListener)

    def unregisterEventListener(self, subscriberid, subscriptionName, listenerKey):
        listenerid = self.event_service.unregisterEventListener(subscriberid, subscriptionName, listenerKey)
        self._removeEventListener(listenerid)

    def removeSubscription(self, subscriberid, subscriptionName):
        listenerList = self.event_service.removeSubscription(subscriberid, subscriptionName)
        for listenerid in listenerList:
            self._removeEventListener(listenerid)

    def _addEventListener(self, listenerid, refListener):
        if listenerid not in self.listenerDirectory:
            self.listenerDirectory[listenerid] = refListener

    def _removeEventListener(self, listenerid):
        if listenerid in self.listenerDirectory:
            del self.listenerDirectory[listenerid]

    # TODO: To be deleted#
    def _processEvent(self, listenerid, topicName, theEvent):
        self.event_service._processEvent(listenerid, topicName, theEvent)


# TODO: Is services.py the right placeholder for this class?#
class EventServiceCmpProxy(EventServiceProxy):
    def __init__(self, service_proxy):
        self.service_proxy = service_proxy
        self.listenerDirectory = {}

    def getTopic(self, topicName):
        msg_id = self.service_proxy._invoke_service(self.service_proxy.fwk.component_id,
                                                    'getTopic', topicName)
        return self.service_proxy._get_service_response(msg_id, True)

    def existsTopic(self, topicName):
        msg_id = self.service_proxy._invoke_service(self.service_proxy.fwk.component_id,
                                                    'existsTopic', topicName)
        return self.service_proxy._get_service_response(msg_id, True)

    def registerSubscriber(self):
        msg_id = self.service_proxy._invoke_service(self.service_proxy.fwk.component_id,
                                                    'registerSubscriber')
        return self.service_proxy._get_service_response(msg_id, True)

    def unregisterSubscriber(self, subscriberid):
        msg_id = self.service_proxy._invoke_service(self.service_proxy.fwk.component_id,
                                                    'unregisterSubscriber', subscriberid)
        listenerList = self.service_proxy._get_service_response(msg_id, True)
        for listenerid in listenerList:
            self._removeEventListener(listenerid)

    def getSubscription(self, subscriberid, subscriptionName):
        msg_id = self.service_proxy._invoke_service(self.service_proxy.fwk.component_id,
                                                    'getSubscription', subscriberid, subscriptionName)
        self.service_proxy._get_service_response(msg_id, True)

    def processEvents(self, subscriberid):
        msg_id = self.service_proxy._invoke_service(self.service_proxy.fwk.component_id,
                                                    'processEvents', subscriberid)
        eventList = self.service_proxy._get_service_response(msg_id, True)
        for listenerid in list(eventList.keys()):
            for topicName in list(eventList[listenerid].keys()):
                for theEvent in eventList[listenerid][topicName]:
                    self.listenerDirectory[listenerid].processEvent(topicName, theEvent)

    def sendEvent(self, topicName, eventName, eventBody):
        msg_id = self.service_proxy._invoke_service(self.service_proxy.fwk.component_id,
                                                    'sendEvent', topicName, eventName, eventBody)
        self.service_proxy._get_service_response(msg_id, True)

    def createListener(self):
        msg_id = self.service_proxy._invoke_service(self.service_proxy.fwk.component_id,
                                                    'createListener')
        return self.service_proxy._get_service_response(msg_id, True)

    def registerEventListener(self, subscriberid, subscriptionName, listenerKey, listenerid, refListener):
        msg_id = self.service_proxy._invoke_service(self.service_proxy.fwk.component_id,
                                                    'registerEventListener', subscriberid, subscriptionName, listenerKey, listenerid)
        self.service_proxy._get_service_response(msg_id, True)
        self._addEventListener(listenerid, refListener)

    def unregisterEventListener(self, subscriberid, subscriptionName, listenerKey):
        msg_id = self.service_proxy._invoke_service(self.service_proxy.fwk.component_id,
                                                    'unregisterEventListener', subscriberid, subscriptionName, listenerKey)
        listenerid = self.service_proxy._get_service_response(msg_id, True)
        self._removeEventListener(listenerid)

    def removeSubscription(self, subscriberid, subscriptionName):
        msg_id = self.service_proxy._invoke_service(self.service_proxy.fwk.component_id,
                                                    'removeSubscription', subscriberid, subscriptionName)
        listenerList = self.service_proxy._get_service_response(msg_id, True)
        for listenerid in listenerList:
            self._removeEventListener(listenerid)

    def _addEventListener(self, listenerid, refListener):
        if listenerid not in self.listenerDirectory:
            self.listenerDirectory[listenerid] = refListener

    def _removeEventListener(self, listenerid):
        if listenerid in self.listenerDirectory:
            del self.listenerDirectory[listenerid]

    # TODO: To be deleted#
    def _processEvent(self, listenerid, topicName, theEvent):
        msg_id = self.service_proxy._invoke_service(self.service_proxy.fwk.component_id,
                                                    '_processEvent', listenerid, topicName, theEvent)
        self.service_proxy._get_service_response(msg_id, True)



# -------------------------------------------------------------------------------
# Copyright 2006-2020 UT-Battelle, LLC. See LICENSE for more information.
# -------------------------------------------------------------------------------
import os
from math import ceil
from . import messages, configurationManager
from .ipsExceptions import BlockedMessageException, \
    IncompleteCallException, \
    InsufficientResourcesException, \
    BadResourceRequestException, \
    AllocatedNodeDownException, \
    ResourceRequestMismatchException
from .ipsutil import which

# from event_service_spec import PublisherEventService,SubscriberEventService,EventListener,Topic,EventServiceException


class TaskManager:
    """
    The task manager is responsible for facilitating component method
    invocations, and the launching of tasks.
    """
    # TM __init__

    def __init__(self, fwk):
        # ref to framework
        self.fwk = fwk
        self.event_mgr = None
        self.data_mgr = None
        self.resource_mgr = None
        self.config_mgr = None
        self.comp_registry = configurationManager.ComponentRegistry()
        self.service_methods = ['init_call',
                                'launch_task',
                                # 'launchTask',   --- deprecated
                                'wait_call',
                                'init_task',
                                'init_task_pool',
                                'finish_task']
        # **** this si where service methods are registered
        self.fwk.register_service_handler(self.service_methods,
                                          getattr(self, 'process_service_request'))
        self.task_map = {}
        self.task_launch_cmd = ''

        # table of currently running tasks
        self.curr_task_table = {}
        # nextCall
        self.next_call_id = 1
        self.next_task_id = 1
        self.outstanding_calls = {}
        self.finished_calls = {}
        self.mpicmd = None  # USed only for CCM on edison

    # this is where messages are received and then something smart happens
    def process_service_request(self, msg):
        """
        Invokes the appropriate public data manager method for the component
        specified in *msg*.  Return method's return value.
        """
        self.fwk.debug('Task Manager received message: %s', str(msg.__dict__))
        method = getattr(self, msg.target_method)
        retval = method(msg)
        return retval

    # TM initialize
    def initialize(self, data_mgr, resource_mgr, config_mgr):
        """
        Initialize references to other managers and key values from
        configuration manager.
        """
        self.event_mgr = None  # eventManager(self)
        self.data_mgr = data_mgr
        self.resource_mgr = resource_mgr
        self.config_mgr = config_mgr
        self.host = self.config_mgr.get_platform_parameter('HOST')
        self.node_alloc_mode = self.config_mgr.get_platform_parameter('NODE_ALLOCATION_MODE')
        try:
            self.task_launch_cmd = self.config_mgr.get_platform_parameter('MPIRUN')
        except Exception:
            print('Error accessing platform parameter MPIRUN')
            raise

        # do later - subscribe to events, set up event publishing structure
        # publish "TM initialized" event

    def get_call_id(self):
        """
        Return a new call id
        """
        retval = self.next_call_id
        self.next_call_id = self.next_call_id + 1
        return retval

    def get_task_id(self):
        """
        Return a new task id
        """
        retval = self.next_task_id
        self.next_task_id = self.next_task_id + 1
        return retval

    def printCurrTaskTable(self):
        """
        Prints the task table pretty-like.
        """
        ctt = self.curr_task_table
        for c, i in list(ctt.items()):
            print(c)
            for k, v in list(i.items()):
                print("   ", k, "=", v)
            print("------")
        print("=====================")

    # TM call
    def init_call(self, init_call_msg, manage_return=True):
        r"""
        Creates and sends a :py:obj:`messages.MethodInvokeMessage` from
        the calling component
        to the target component.  If *manage_return* is ``True``, a record is
        added to *outstanding_calls*.  Return call id.

        Message args:

          0. method_name

          1. \+ arguments to be passed on as method arguments.
        """
        callee_id = init_call_msg.target_comp_id
        method_name = init_call_msg.args[0]
        args = init_call_msg.args[1:]
        keywords = init_call_msg.keywords
        caller_id = init_call_msg.sender_id
        call_id = self.get_call_id()
        self.fwk.debug('TM:init_call(): %s %s %s %s',
                       caller_id, callee_id, method_name, str(args))
        invoke_msg = messages.MethodInvokeMessage(self.fwk.component_id,
                                                  callee_id,
                                                  call_id,
                                                  method_name, *args, **keywords)
        invocation_q = self.comp_registry.getComponentArtifact(callee_id,
                                                               'invocation_q')
        invocation_q.put(invoke_msg)
        if manage_return:
            self.outstanding_calls[call_id] = (caller_id, None)
        return call_id

    def return_call(self, response_msg):
        """
        Handle the response message generated by a component in response
        to a method invocation on that component.

        *reponse_msg* is expected to be of type :py:obj:`messages.MethodResultMessage`
        """
        call_id = response_msg.call_id
        caller_id = self.outstanding_calls[call_id][0]
        self.fwk.debug('TM:call_return() call_id = %s caller_id = %s', call_id, caller_id)
        self.finished_calls[call_id] = (caller_id, response_msg)
        del self.outstanding_calls[call_id]

    def wait_call(self, wait_msg):
        """
        Determine if the call has finished.  If finished, return any data or
        errors.  If not finished raise the appropriate blocking or nonblocking
        exception and try again later.

        *wait_msg* is expected to be of type :py:obj:`messages.ServiceRequestMessage`

        Message args:

        0. *call_id*: call id for which to wait

        1. *blocking*: determines the wait is blocking or not
        """
        call_id = wait_msg.args[0]
        blocking = wait_msg.args[1]
        self.fwk.debug('TM:wait_call() call_id = %s', call_id)
        if call_id in self.finished_calls:
            response_msg = self.finished_calls[call_id][1]
            del self.finished_calls[call_id]
            if response_msg.status == messages.Message.FAILURE:
                raise Exception(response_msg.args[0])
            else:
                return response_msg.args
        if not blocking:
            raise IncompleteCallException(call_id)
        else:
            raise BlockedMessageException(wait_msg, '***call %s not finished' % call_id)

    def init_task(self, init_task_msg):
        r"""
        Allocate resources needed for a new task and build the task
        launch command using the binary and arguments provided by
        the requesting component.  Return launch command to component via
        :py:obj:`messages.ServiceResponseMessage`.  Raise exception if task
        can not be launched at this time (:py:exc:`ipsExceptions.BadResourceRequestException`, :py:exc:`ipsExceptions.InsufficientResourcesException`).

        *init_task_msg* is expected to be of type :py:obj:`messages.ServiceRequestMessage`

        Message args:

        0. *nproc*: number of processes the task needs

        1. *binary*: full path to the executable to launch

        # SIMYAN: added this to deal with the component directory change
        2. *working_dir*: full path to directory where the task will be launched

        3. *tppn*: processes per node for this task.  (0 indicates that the default ppn is used.)

        4. *block*: whether or not to wait until the task can be launched.

        5. *wnodes*: ``True`` for whole node allocation, ``False`` otherwise.

        6. *wsocks*: ``True`` for whole socket allocation, ``False`` otherwise.

        7. \+ *cmd_args*: any arguments for the executable
        """
        caller_id = init_task_msg.sender_id
        nproc = int(init_task_msg.args[0])
        binary = init_task_msg.args[1]
        # SIMYAN: working_dir stored
        working_dir = init_task_msg.args[2]
        tppn = int(init_task_msg.args[3])  # task processes per node
        block = init_task_msg.args[4]  # Block waiting for available resources
        wnodes = init_task_msg.args[5]
        wsocks = init_task_msg.args[6]

        # SIMYAN: increased arguments
        cmd_args = init_task_msg.args[7:]
        # handle for task related things
        task_id = self.get_task_id()

        try:
            retval = self.resource_mgr.get_allocation(caller_id,
                                                      nproc,
                                                      task_id,
                                                      wnodes,
                                                      wsocks,
                                                      task_ppn=tppn)
            self.fwk.debug('RM: get_allocation() returned %s', str(retval))
            partial_node = retval[0]
            if partial_node:
                (nodelist, corelist, ppn, max_ppn, accurateNodes) = retval[1:]
            else:
                (nodelist, ppn, max_ppn, accurateNodes) = retval[1:]
        except InsufficientResourcesException:
            if block:
                raise BlockedMessageException(init_task_msg, '***%s waiting for %d resources' %
                                              (caller_id, nproc))
            else:
                raise
        except BadResourceRequestException as e:
            self.fwk.error("There has been a fatal error, %s requested %d too many processors in task %d",
                           caller_id, e.deficit, e.task_id)
            raise
        except ResourceRequestMismatchException as e:
            self.fwk.error("There has been a fatal error, %s requested too few processors per node to launch task %d (requested: procs = %d, ppn = %d)",
                           caller_id, e.task_id, e.nproc, e.ppn)
            raise
        except Exception:
            raise

        # SIMYAN: moved up a few lines and ret_data, node_file added
        self.curr_task_table[task_id] = {'component': caller_id,
                                         'status': 'init_task',
                                         'binary': binary,
                                         'nproc': nproc,
                                         'args': cmd_args,
                                         'launch_cmd': None,
                                         'env_update': None,
                                         'ret_data': None,
                                         'node_file': None}
        if partial_node:
            nodes = ','.join(nodelist)
            (cmd, env_update) = self.build_launch_cmd(nproc, binary, cmd_args,
                                                      working_dir, ppn,
                                                      max_ppn, nodes,
                                                      accurateNodes,
                                                      partial_node, task_id,
                                                      core_list=corelist)
        else:
            if accurateNodes:
                nodes = ','.join(nodelist)
            else:
                nodes = ''
            (cmd, env_update) = self.build_launch_cmd(nproc, binary, cmd_args,
                                                      working_dir, ppn,
                                                      max_ppn, nodes,
                                                      accurateNodes,
                                                      False, task_id)
        task_data = self.curr_task_table[task_id]
        task_data['launch_cmd'] = cmd
        task_data['env_update'] = env_update
        return (task_id, cmd, env_update)

    def build_launch_cmd(self, nproc, binary, cmd_args, working_dir, ppn,
                         max_ppn, nodes, accurateNodes, partial_nodes,
                         task_id, core_list=''):
        """
        Construct task launch command to be executed by the component.

         * nproc - number of processes to use
         * binary - binary to launch
         * cmd_args - additional command line arguments for the binary
         * working_dir - full path to where the executable will be launched
         * ppn - processes per node value to use
         * max_ppn - maximum possible ppn for this allocation
         * nodes - comma separated list of node ids
         * accurateNodes - if ``True``, launch on nodes in *nodes*, otherwise the parallel launcher determines the process placement
         * partial_nodes - if ``True`` and *accurateNodes* and *task_launch_cmd* == 'mpirun',
               a host file is created specifying the exact placement of processes on cores.
         * core_list - used for creating host file with process to core mappings
        """
        # set up launch command
        env_update = None
        nproc_flag = ''
        smp_node = len(self.resource_mgr.nodes) == 1

        if self.task_launch_cmd == 'eval':
            # cmd = binary
            if len(cmd_args) > 0:
                cmd_args = ' '.join(cmd_args)
                cmd = ' '.join([binary, cmd_args])
            else:
                cmd = binary
            return cmd, env_update

        # -------------------------------------
        # mpirun
        # -------------------------------------
        elif self.task_launch_cmd == 'mpirun':
            version = self.config_mgr.get_platform_parameter('MPIRUN_VERSION').upper()
            if version.startswith("OPENMPI"):
                if version == 'OPENMPI-DVM':
                    mpi_binary = 'prun'
                    smp_node = False
                else:  # VERSION = OPENMPI_GENERIC
                    mpi_binary = 'mpirun'
                # Find and cache full path to launch executable
                if not self.mpicmd:
                    self.mpicmd = which(mpi_binary)
                mpicmd = self.mpicmd
                if not mpicmd:
                    raise Exception('Missing %s command in $PATH' % (mpi_binary))

                nproc_flag = '-np'
                ppn_flag = '-npernode'
                host_select = '-H'
                if smp_node or mpi_binary == 'prun':
                    cmd = ' '.join([mpicmd,
                                    nproc_flag, str(nproc)])
                else:
                    cmd = ' '.join([mpicmd,
                                    nproc_flag, str(nproc),
                                    ppn_flag, str(ppn)])
                cmd = f"{cmd} -x PYTHONPATH"  # Propagate PYTHONPATH to compute nodes
                if accurateNodes:
                    cmd = ' '.join([cmd, host_select, nodes])
            elif version == 'SGI':
                if accurateNodes:
                    core_dict = {}
                    ppn_groups = {}
                    num_cores = self.resource_mgr.cores_per_socket
                    for (n, cl) in core_list:
                        core_dict.update({n: cl})
                        if len(cl) in list(ppn_groups.keys()):
                            ppn_groups[len(cl)].append(n)
                        else:
                            ppn_groups.update({len(cl): [n]})
                    cmdlets = []
                    envlets = []
                    bin_n_args = binary + ' '.join(cmd_args)
                    for p, ns in list(ppn_groups.items()):
                        cmdlets.append(' '.join([','.join(ns), str(p),
                                                 bin_n_args]))
                        el_node = []
                        for n in ns:
                            el_tmp = []
                            for k in core_dict[n]:
                                s, c = k.split(':')
                                s = int(s)
                                c = int(c)
                                el_tmp.append(str(s * num_cores + c))
                            el_node.append(','.join(el_tmp))
                        envlets.append(':'.join(el_node))
                    cmd = self.task_launch_cmd + ' ' + ' : '.join(cmdlets)
                    env_update = {'MPI_DSM_CPULIST': ':'.join(envlets)}
                    return cmd, env_update
                else:
                    cmd = ' '.join([self.task_launch_cmd, ppn, binary,
                                    ' '.join(cmd_args)])

        # --------------------------------------
        # mpiexec (MPICH variants)
        # --------------------------------------
        elif self.task_launch_cmd == 'mpiexec':
            nproc_flag = '-n'
            ppn_flag = '-npernode'
            if smp_node:
                cmd = ' '.join([self.task_launch_cmd, nproc_flag, str(nproc)])
            elif self.host == 'iter':
                cfg_fname = ".node_config_" + str(task_id)
                cfg_fname = os.path.join(working_dir, cfg_fname)
                cfg_file = open(cfg_fname, 'w')
                cmd_args = ' '.join(cmd_args)
                node_command = ' '.join([binary, cmd_args])
                node_spec = ''
                if partial_nodes:
                    for (node, cores) in core_list:
                        node_spec += ('%s ' % (node)) * len(cores)
                else:
                    for node in nodes.split(' ,'):
                        node_spec += ('%s ' % (node)) * ppn
                print('%s: %s' % (node_spec, node_command), file=cfg_file)
                config_option = '-config=' + cfg_fname
                cmd = ' '.join([self.task_launch_cmd, config_option])
                self.curr_task_table[task_id]['node_file'] = cfg_fname
                return cmd, env_update
            elif accurateNodes:  # Need to assign tasks to nodes explicitly
                host_select = '--host ' + nodes
                cmd = ' '.join([self.task_launch_cmd, host_select,
                                nproc_flag, str(nproc), ppn_flag,
                                str(ppn)])
            else:
                cmd = ' '.join([self.task_launch_cmd, nproc_flag,
                                str(nproc), ppn_flag, str(ppn)])
        # ------------------------------------
        # aprun (Cray parallel launch)
        # ------------------------------------
        elif self.task_launch_cmd == 'aprun':
            nproc_flag = '-n'
            ppn_flag = '-N'
            cpu_assign_flag = '-cc'
            by_numanode_flag = '-S'
            if self.host in ['hopper', 'edison']:
                num_numanodes = self.resource_mgr.sockets_per_node
                num_cores = self.resource_mgr.cores_per_node
                if accurateNodes:
                    nlist_flag = '-L'
                    num_nodes = len(nodes.split(','))
                    ppn = int(ceil(float(nproc) / num_nodes))
                    per_numa = int(ceil(float(ppn) / num_numanodes))
                    if per_numa == num_cores / num_numanodes:

                        cmd = ' '.join([self.task_launch_cmd,
                                        nproc_flag, str(nproc),
                                        ppn_flag, str(ppn),
                                        nlist_flag, nodes])
                    else:
                        if num_nodes > 1:
                            ppn = per_numa * num_numanodes
                        if nproc < ppn:
                            ppn = nproc
                        cmd = ' '.join([self.task_launch_cmd,
                                        nproc_flag, str(nproc),
                                        ppn_flag, str(ppn),
                                        by_numanode_flag, str(per_numa),
                                        nlist_flag, nodes])
                else:
                    num_nodes = int(ceil(float(nproc) / ppn))
                    ppn = int(ceil(float(nproc) / num_nodes))
                    per_numa = int(ceil(float(ppn) / num_numanodes))
                    if per_numa == self.resource_mgr.cores_per_node / self.resource_mgr.sockets_per_node:

                        cmd = ' '.join([self.task_launch_cmd,
                                        nproc_flag, str(nproc),
                                        ppn_flag, str(ppn)])
                    else:
                        if num_nodes > 1:
                            ppn = per_numa * num_numanodes
                        if nproc < ppn:
                            ppn = nproc
                        cmd = ' '.join([self.task_launch_cmd,
                                        nproc_flag, str(nproc),
                                        ppn_flag, str(ppn),
                                        by_numanode_flag, str(per_numa)])
            else:
                if accurateNodes:
                    nlist_flag = '-L'
                    cmd = ' '.join([self.task_launch_cmd,
                                    nproc_flag, str(nproc),
                                    ppn_flag, str(ppn),
                                    nlist_flag, nodes])
                else:
                    cmd = ' '.join([self.task_launch_cmd,
                                    nproc_flag, str(nproc),
                                    cpu_assign_flag,
                                    '%d-%d' % (max_ppn - 1, max_ppn - int(ppn)),
                                    ppn_flag, str(ppn)])
        # ------------------------------------
        # numactl (single process launcher)
        # ------------------------------------
        elif self.task_launch_cmd == 'numactl':
            if accurateNodes and partial_nodes:
                proc_flag = '--physcpubind='
                procs = ''
                for p in core_list:
                    procs = ','.join([k.split(':')[1] for k in p[1]])
                proc_flag += procs
            else:
                self.fwk.warning('numactl needs accurateNodes')
                proc_flag = ''
            cmd = ' '.join([self.task_launch_cmd,
                            proc_flag])
        elif self.task_launch_cmd == 'srun':
            nproc_flag = '-n'
            nnodes_flag = '-N'
            num_nodes = len(nodes.split(','))
            cmd = ' '.join([self.task_launch_cmd, nnodes_flag,
                            str(num_nodes), nproc_flag, str(nproc)])
        else:
            self.fwk.exception("invalid task launch command.")
            raise "invalid task launch command."

        cmd_args = ' '.join(cmd_args)
        cmd = ' '.join([cmd, binary, cmd_args])

        return cmd, env_update

    def init_task_pool(self, init_task_msg):
        """
        Allocate resources needed for a new task and build the task
        launch command using the binary and arguments provided by
        the requesting component.

        *init_task_msg* is expected to be of type :py:obj:`messages.ServiceRequestMessage`

        Message args:

        0. *task_dict*: dictionary of task names and objects
        """
        caller_id = init_task_msg.sender_id
        task_dict = init_task_msg.args[0]
        ret_dict = {}
        for task_name in list(task_dict.keys()):
            # handle for task related things
            task_id = self.get_task_id()
            (nproc, working_dir, binary, cmd_args, tppn, wnodes, wsocks) = task_dict[task_name]

            try:
                retval = self.resource_mgr.get_allocation(caller_id,
                                                          nproc,
                                                          task_id,
                                                          wnodes, wsocks,
                                                          task_ppn=tppn)
                self.fwk.debug('RM: get_allocation() returned %s', str(retval))
                partial_node = retval[0]
                if partial_node:
                    (nodelist, corelist, ppn, max_ppn, accurateNodes) = retval[1:]
                else:
                    (nodelist, ppn, max_ppn, accurateNodes) = retval[1:]

            except InsufficientResourcesException:
                continue
            except BadResourceRequestException as e:
                self.fwk.error("There has been a fatal error, %s requested %d too many processors in task %d",
                               caller_id, e.deficit, e.task_id)
                for (task_id, cmd) in list(ret_dict.values()):
                    self.resource_mgr.release_allocation(task_id, -1)
                    del self.curr_task_table[task_id]
                raise
            except ResourceRequestMismatchException as e:
                self.fwk.error("There has been a fatal error, %s requested too few processors per node to launch task %d (request: procs = %d, ppn = %d)",
                               caller_id, e.task_id, e.nproc, e.ppn)
                for (task_id, cmd) in list(ret_dict.values()):
                    self.resource_mgr.release_allocation(task_id, -1)
                    del self.curr_task_table[task_id]
                raise
            except Exception:
                self.fwk.exception('TM:init_task_pool(): Allocation exception')
                raise

            if partial_node:
                nodes = ','.join(nodelist)
                (cmd, env_update) = self.build_launch_cmd(nproc, binary,
                                                          cmd_args,
                                                          working_dir, ppn,
                                                          max_ppn, nodes,
                                                          accurateNodes,
                                                          partial_node,
                                                          task_id,
                                                          core_list=corelist)
            else:
                if accurateNodes:
                    nodes = ','.join(nodelist)
                else:
                    nodes = ''
                (cmd, env_update) = self.build_launch_cmd(nproc, binary,
                                                          cmd_args,
                                                          working_dir,
                                                          ppn, max_ppn, nodes,
                                                          accurateNodes, False,
                                                          task_id)

            self.curr_task_table[task_id] = {'component': caller_id,
                                             'status': 'init_task',
                                             'binary': binary,
                                             'nproc': nproc,
                                             'args': cmd_args,
                                             'launch_cmd': cmd,
                                             'env_update': env_update,
                                             'ret_data': None}
            ret_dict[task_name] = (task_id, cmd, env_update)
        return ret_dict

    def finish_task(self, finish_task_msg):
        """
        Cleanup after a task launched by a component terminates

        *finish_task_msg* is expected to be of type :py:obj:`messages.ServiceRequestMessage`

        Message args:

        0. *task_id*: task id of finished task

        1. *task_data*: return code of task
        """
        task_id = finish_task_msg.args[0]
        task_data = finish_task_msg.args[1]
        try:
            self.resource_mgr.release_allocation(task_id, task_data)
            del self.curr_task_table[task_id]
        except AllocatedNodeDownException:
            del self.curr_task_table[task_id]
            return 1
        except Exception:
            print('Error finishing task ', task_id)
            raise
        return 0
